# coding: utf-8

import numpy as np

def index_max(lst):
  return lst.index(max(lst))

# used for one to many mapping
class multi_dict(dict):
  def add(self, key, item):
    if key in self: self[key].append(item)
    else:           self[key] = [ item ]

  # performs __getitem__ and if the resulting list has only one
  # element, returns that element not in a list
  def get_flattened(self, key):
    result = self[key]
    # if there is only one item for that key, only return that
    # item. otherwise, return a set containing those items
    if len(result) == 1: return result[0]
    else:                return result

class counting_dict(dict):
  def add(self, item):
    if item in self: self[item] += 1
    else:            self[item] = 0

class hashable_dict:
  def __init__(self, d):
    self.my_dict = d
    self.my_frozenset = frozenset(list(d.items()))
  def __getitem__(self, item):
    return self.my_dict[item]
  def __hash__(self):
    return hash(self.my_frozenset)
  def __eq__(self, rhs):
    return isinstance(rhs, hashable_dict) and self.my_frozenset == rhs.my_frozenset
  def __ne__(self, rhs):
    return not self == rhs
  def __str__(self):
    return 'hashable_dict(' + str(self.my_dict) + ')'
  def __repr__(self):
    return self.__str__()

def log_add(logA,logB):
  if logA == None: return logB
  if logA < logB:  return log_add(logB,logA)
  return np.log2( 1 + pow(2, logB-logA) ) + logA

def log_sum(lst):
  log_total = None
  for v in lst: log_total = log_add(log_total, v)
  return log_total

def prod(lst):
  result = 1.0
  for i in lst: result *= i
  return result

def binomial(n, r):
  ### n! /( r! (n-r)! )
  return prod(list(range(r+1, n+1))) / prod(2, n-r+1)




#!/usr/bin/env python3
# coding: utf-8

import argparse
import logging
import numpy as np
import os
import pandas as pd
import time

from dart_id.align import align
from dart_id.converter import process_files
from dart_id.exceptions import ConfigFileError
from dart_id.fido.BayesianNetwork import run_internal
from dart_id.figures import figures
from dart_id.models import models, get_model_from_config
from dart_id.helper import add_global_args, read_config_file, init_logger, load_params_from_file
from scipy.stats import norm, lognorm, laplace, bernoulli, uniform

logger = logging.getLogger('root')

def update(dfa, params, config):
  dfa = dfa.reset_index(drop=True)

  #logger.info('{} / {} ({:.2%}) confident, alignable observations (PSMs) after filtering.'.format(dff.shape[0], dfa.shape[0], dff.shape[0] / dfa.shape[0]))

  # refactorize peptide id into stan_peptide_id, 
  # to preserve continuity when feeding data into STAN
  dfa['stan_peptide_id'] = dfa['sequence'].map({
    ind: val for val, ind in enumerate(dfa['sequence'].unique())})

  num_experiments = dfa['exp_id'].max() + 1
  num_peptides = dfa['peptide_id'].max() + 1
  exp_names = np.sort(dfa['raw_file'].unique())
  pep_id_list = dfa['peptide_id'].unique()

  # validate parameters file. make sure it is from the same filters
  # or else the program will crash in the code below
  # check num_experiments, num_peptides
  if params['exp'].shape[0] != num_experiments or \
     params['peptide'].shape[0] != (dfa['stan_peptide_id'].max() + 1):
    raise ConfigFileError('Parameters files have different data than the input data provided. Ensure that both the input list and filters used to generate the alignment parameters and those provided to the current update are the __exact__ same.')
  
  model = get_model_from_config(config)

  # mu from the STAN alignment
  dfa['mu'] = params['peptide']['mu'].values[dfa['stan_peptide_id']] 
  
  # concatenate transformation parameters
  exp_params = pd.DataFrame({ key: params['exp'][key][dfa['exp_id']] \
    for key in model['exp_keys']}).reset_index(drop=True)
  dfa = pd.concat([dfa, exp_params], axis=1)

  # predict mus with RTs, and RTs with aligned mus
  dfa['mu_pred'] = model['rt_to_ref'](dfa, dfa['mu'], params)
  dfa['muij']    = model['ref_to_rt'](dfa, dfa['mu'], params)
  dfa['sigmaij'] = model['sigmaij_func'](dfa, params)
  # scaled sigma is the same ratio of muij / mu applied to sigmaij
  dfa['sigma_pred'] = dfa['sigmaij'] * dfa['mu_pred'] / dfa['muij']

  # get parameters for the null distributions for each experiment
  null_dists = dfa.groupby('exp_id')['retention_time'].agg([np.mean, np.std])
  #null_dists = np.array([norm(loc=null_dists.loc[i, 'mean'], scale=null_dists.loc[i, 'std']) for i in range(0, num_experiments)])
  # first column is mean, second is std
  null_dists = np.array([null_dists['mean'].values, null_dists['std'].values]).T

  # PEP ceiling at 1, otherwise will result in 
  # incorrect negative densities when plugging into Bayes' theorem
  dfa['pep'][dfa['pep'] > 1.0] = 1.0

  # output table
  df_new = pd.DataFrame()

  bootstrap_method = 'none'
  if 'bootstrap_method' in config:
    bootstrap_method = config['bootstrap_method']
    logger.info('Using \"{}\" bootstrap method'.format(bootstrap_method))
  else:
    logger.info('Bootstrap method not defined, using point estimates to update confidence instead.')

  k = 20 # default
  if 'bootstrap_iters' in config:
    k = config['bootstrap_iters']
    if bootstrap_method != 'none':
      logger.info('Using {} bootstrap iterations'.format(k))

  logger.info('Updating PEPs...')
  for i, e in enumerate(np.sort(dfa['exp_id'].unique())):

    exp_name = exp_names[i]

    exp = dfa[dfa['exp_id'] == e]
    exp = exp.reset_index(drop=True)

    exp_peptides = exp['stan_peptide_id'].unique()

    logger.info('Exp ({} / {}) - {} - ({} Peptides, {} PSMs)'.format(i+1, num_experiments, exp_name, len(exp_peptides), exp.shape[0]))

    # vector of P(RT|delta=1) for this experiment.
    rt_plus = pd.Series(np.zeros(exp.shape[0]))

    if bootstrap_method != 'none':

      # to avoid using this experiment's own data to update the confidence
      # of its own observations, recalculate the reference RTs (mu) without the
      # data from this experiment, by: 
      # 1) non-parametric bootstrapping over the median of the predicted mus.
      # OR
      # 2) parametric bootstrapping, using the RT distribution parameters
      
      # get predicted mus of peptides in this experiment, excluding predicted mus
      # transformed from RTs observed in this experiment
      dfe = dfa.loc[((dfa['stan_peptide_id'].isin(exp_peptides)) & (dfa['exp_id'] != e)), \
        ['stan_peptide_id', 'pep', 'mu_pred', 'mu', 'sigma_pred', 'exp_id']]

      # extract relevant values for each peptide
      mu_preds    = dfe.groupby('stan_peptide_id')['mu_pred'].apply(lambda x: x.values).values.tolist()
      mus         = dfe.groupby('stan_peptide_id')['mu'].apply(lambda x: x.values).values.tolist()
      sigma_preds = dfe.groupby('stan_peptide_id')['sigma_pred'].apply(lambda x: x.values).values.tolist()
      peps        = dfe.groupby('stan_peptide_id')['pep'].apply(lambda x: x.values).values.tolist()
      exp_ids     = dfe.groupby('stan_peptide_id')['exp_id'].apply(lambda x: x.values).values.tolist()
      
      # number of observations per peptide sequence
      obs_per_seq = [len(peptide) for peptide in mu_preds]
      num_peptides = len(mu_preds)

      # the number of observations per peptide -- used in loop
      num_obs = 0
      # matrix of n by k estimated mus from the bootstrapping
      # will iterate over in the loop after the immediate one
      mu_k = np.zeros((num_peptides, k))

      if bootstrap_method == 'parametric' or bootstrap_method == 'parametric_mixture':

        t_laplace_samples = 0
        t_coin_flips = 0
        t_null_samples = 0
        t_loop_indexing = 0
        t_medians = 0

        # create pool of coin flips, instead of sampling for every peptide
        # the pool is uniformly distributed from 0 to 1, and "successful" coin flip
        # is determined by whether or not the sample from the pool is less than the
        # measured PEP
        _time = time.time()
        coin_flip_pool = 0
        if bootstrap_method == 'parametric_mixture':
          coin_flip_pool = uniform.rvs(size=(np.sum(obs_per_seq) * k))

        t_coin_flips += (time.time() - _time)
        coin_counter = 0

        # create a pool of laplace samples, to pull from for each peptide
        _time = time.time()
        sample_pool = laplace.rvs(size=(np.sum(obs_per_seq) * k))
        t_laplace_samples += (time.time() - _time)
        # keep track of where we are in the pool with a counter
        sample_counter = 0

        # parametric bootstrap
        for i in range(0, num_peptides):
          num_obs = obs_per_seq[i]
          
          _time = time.time()
          # sample num_obs synthetic RTs for k bootstrap iterations
          # do the sampling in a big pool, then shape to matrix where
          # rows correspond to bootstrap iters and columns correspond to sample observations
          #samples = laplace.rvs(size=(k * num_obs)).reshape(k, num_obs)
          
          # draw samples from sample pool, reshape into matrix
          _time = time.time()
          samples = sample_pool[sample_counter:(sample_counter+(k*num_obs))]
          samples = samples.reshape(k, num_obs)

          # increment sample counter
          sample_counter += (k*num_obs)

          #mu_med = np.median()
          # shift and scale sampled RTs by mu and sigma_pred, respectively
          samples = (samples * sigma_preds[i]) + mu_preds[i]
          #samples = (samples * sigma_preds[i]) + mu_med
          t_laplace_samples += (time.time() - _time)

          if bootstrap_method == 'parametric_mixture':
            # sample from mixture distribution
            _time = time.time()
            # actually faster to just replicate the sample matrix and then
            # take subindices from that instead of sampling from null every
            # iteration of the loop below. this seems inefficient, especially
            # if given very small PEPs, but still better than sampling every iteration.
            # could probably optimize the size of the null sample matrix by
            # looking at predicted false positive rates, but for now we're
            # just going with worst case scenario and assuming for all false positives.
            null_samples = norm.rvs(size=(k * num_obs)).reshape(k, num_obs)
            # shift and scale sampled RTs by mean and std of null dists
            null_samples = (null_samples * null_dists[exp_ids[i],1]) + null_dists[exp_ids[i],0]
            t_null_samples += (time.time() - _time)

            _time = time.time()
            for j in range(0, num_obs): # for each observation in the matrix
              # take a chunk of the coin flip pool
              fp = (coin_flip_pool[coin_counter:(coin_counter + k)] < peps[i][j]).astype(bool)
              coin_counter += k
              # overwrite original samples with samples from null distribution
              samples[fp, j] = null_samples[fp, j]
            t_loop_indexing += (time.time() - _time)

          _time = time.time()
          # now take the median of each row and store it in mu_k
          mu_k[i] = np.median(samples, axis=1)
          # or take the weighted mean
          #weights = ((1 - peps[i]) - (1 - config['pep_threshold'])) / config['pep_threshold']
          #mu_k[i] = (np.sum(samples * weights, axis=1) / np.sum(weights))
          t_medians += (time.time() - _time)

        logger.debug('laplace sampling: {:.1f} ms'.format(t_laplace_samples*1000))
        logger.debug('coin flips: {:.1f} ms'.format(t_coin_flips*1000))
        logger.debug('null sampling: {:.1f} ms'.format(t_null_samples*1000))
        logger.debug('loop indexing: {:.1f} ms'.format(t_loop_indexing*1000))
        logger.debug('taking medians: {:.1f} ms'.format(t_medians*1000))

      elif bootstrap_method == 'non-parametric':
        # non-parametric bootstrap
        # instead of generating random indices for the sampling for each
        # iteration, and for each peptide, we'll generate a batch of random numbers
        # now and pull from them later.
        # the counter will keep track of which portion of the pool we're using
        counter = 0
        rand_pool = np.random.rand(np.sum(obs_per_seq) * k)

        for i in range(0, num_peptides): # for each peptide sequence
          num_obs = obs_per_seq[i]
          for j in range(0, k): # for each iteration:
            # re-estimate mu from the resampled mu_preds
            # TODO: choice also of mean, weighted mean
            mu_k[i][j] = np.median(mu_preds[i][\
              (rand_pool[counter:(counter+num_obs)] * num_obs).astype(int, copy=False)])

            counter = counter + num_obs

      else:
        raise ConfigFileError('Invalid bootstrap method. Please choose \"parametric\" or \"non-parametric.\"')
      
      _t_dist_building = time.time()
      # map of stan_peptide_id onto 1:num_peptides
      pep_inds = {ind: var for var, ind in enumerate(exp_peptides)}
      pep_inds = exp['stan_peptide_id'].map(pep_inds)

      # for each bootstrap iteration:
      for j in range(0, k):
        # evaluate the transformed RTs (predicted mus) on distributions
        # with the bootstrapped, estimated mus as the means.
        #rt_plus = rt_plus + laplace.pdf(exp['retention_time'], \
        #  loc=model['ref_to_rt'](exp, mu_k[:,j][pep_inds], params), \
        #  scale=exp['sigmaij'])

        rt_plus = rt_plus + laplace.pdf(exp['mu_pred'], \
          loc=mu_k[:,j][pep_inds], \
          scale=exp['sigma_pred'])

      # divide total likelihood by # of iterations to normalize to area of 1
      rt_plus = rt_plus / k

      logger.debug('distribution building: {:.1f} ms'.format((time.time() - _t_dist_building)*1000))

    else:
      # not using bootstrap, but using adjusted mu as a point estimate
      # for updating the confidence
      rt_plus = model['rt_plus_func'](exp)

    #                                         P(RT|delta=0)*P(delta=0)
    # PEP.new = P(delta=0|RT) =   ---------------------------------------------------
    #                             P(RT|delta=0)*P(delta=0) + P(RT|delta=1)*P(delta=1)
    #                         
    # delta=1 = Correct ID (true positive)
    # delta=0 = Incorrect (false positive)
    
    # P(RT|delta=0) = probability of peptides RT, given that PSM is incorrect
    #           estimate empirical density of RTs over the experiment
    
    rt_minus = model['rt_minus_func'](exp)

    # P(delta=0) = probability that PSM is incorrect (PEP)
    # P(delta=1) = probability that PSM is correct (1-PEP)
    
    # P(RT|delta=1) = probability that given the correct ID, the RT falls in the
    #           normal distribution of RTs for that peptide, for that experiment
      
    # delta=1 = Correct ID (true positive)
    # delta=0 = Incorrect (false positive)
    # 
    pep_new = (rt_minus * exp['pep']) / \
      ((rt_minus * exp['pep']) + (rt_plus * (1.0 - exp['pep'])))

    # for PSMs for which we have alignment/update data
    exp_new = pd.DataFrame({
        'rt_minus':          rt_minus.tolist(),
        'rt_plus':           rt_plus.tolist(),
        'mu':                exp['mu'].values.tolist(),
        'muij':              exp['muij'].values.tolist(),
        'sigmaij':           exp['sigmaij'].values.tolist(),
        'pep_new':           pep_new.tolist(),

        'id':                exp['id'].values,
        'exp_id':            exp['exp_id'].values,
        'peptide_id':        exp['peptide_id'].values,
        'stan_peptide_id':   exp['stan_peptide_id'].values,
        'input_id':          exp['input_id'].values,
        'exclude':           exp['exclude'].values
    })
    # append to master DataFrame and continue
    df_new = df_new.append(exp_new)

  # reorder by ID and reset the index
  df_new = df_new.sort_values('id')
  df_new = df_new.reset_index(drop=True)

  return df_new

def write_output(df, out_path, config):
  # remove diagnostic columns, unless they are specified to be kept
  if 'add_diagnostic_cols' not in config or config['add_diagnostic_cols'] == False:
    df = df.drop(['pep_new', 'participated', 'exclude', 'mu', 'muij', 
      'rt_minus', 'rt_plus', 'sigmaij', 'residual',
      'input_id', 'exp_id', 'peptide_id', 'stan_peptide_id'], axis=1)

  # filter by PSM FDR?
  if 'psm_fdr_threshold' in config and type(config['psm_fdr_threshold']) == float:
    if config['psm_fdr_threshold'] <= 0:
      logger.warning('PSM FDR threshold equal to or below 0. Please provide a value between 0 and 1. Ignoring...')
    elif config['psm_fdr_threshold'] >= 1:
      logger.warning('PSM FDR threshold equal to or greater than 1. Please provide a value between 0 and 1. Ignoring...')
    else:
      to_remove = (df['dart_qval'] > config['psm_fdr_threshold'])
      logger.info('{}/{} ({:.2%}) PSMs removed at a threshold of {:.2%} FDR.'.format(np.sum(to_remove), df.shape[0], np.sum(to_remove) / df.shape[0], config['psm_fdr_threshold']))
      df = df[~to_remove].reset_index(drop=True)

  # filter by protein FDR?
  if 'protein_fdr_threshold' in config and type(config['protein_fdr_threshold']) == float:
    if 'razor_protein_fdr' in df.columns:
      if config['protein_fdr_threshold'] <= 0:
        logger.warning('Protein FDR threshold equal to or below 0. Please provide a value between 0 and 1. Ignoring...')
      elif config['protein_fdr_threshold'] >= 1:
        logger.warning('Protein FDR threshold equal to or greater than 1. Please provide a value between 0 and 1. Ignoring...')
      else:
        to_remove = ((df['razor_protein_fdr'] > config['protein_fdr_threshold']) | pd.isnull(df['razor_protein_fdr']))
        logger.info('{}/{} ({:.2%}) PSMs removed at a threshold of {:.2%} Protein FDR.'.format(np.sum(to_remove), df.shape[0], np.sum(to_remove) / df.shape[0], config['protein_fdr_threshold']))
        df = df[~to_remove].reset_index(drop=True)
    else:
      raise ConfigFileError('Protein FDR threshold specified, but no protein inference run with this analysis. Please set \"run_pi\" to true and fill out all the respective parameters.')
    
  df.to_csv(out_path, sep='\t', index=False)

def main():
  start = time.time()

  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser)
  args = parser.parse_args()

  # load config file
  # this function also creates the output folder
  config = read_config_file(args)

  # initialize logger
  init_logger(config['verbose'], os.path.join(config['output'], 'dart.log'), config['log_file'])

  logger.info('Converting files and filtering PSMs')
  df, df_original = process_files(config)
  logger.info('Finished converting files and filtering PSMs.')

  # load params, either from a defined folder or from running the alignment
  params = {}
  if 'params_folder' in config and type(config['params_folder']) is str:
    params = load_params_from_file(config['params_folder'])
  else:
    logger.info('Beginning alignment procedure')
    params = align(df, config)
    logger.info('Alignment procedure finished')

  # now we have the params, run the update
  logger.info('Updating PEPs with alignment data...')
  df_new = update(df, params, config)

  # save the sparse combined input file?
  #df_new.to_csv(os.path.join(args.output, 'df_converted.txt'), sep='\t', index=False)

  # add new columns to original DF, and remove the duplicate ID column
  logger.info('Concatenating results to original data...')
  df_adjusted = pd.concat([ \
    df_original.loc[~df_original['remove']].reset_index(drop=True), \
    df_new.drop(['id', 'input_id'], axis=1).reset_index(drop=True)], axis=1)

  # add rows of PSMs originally removed from analysis
  if np.sum(df_original['remove']) > 0:
    logger.info('Reattaching {} PSMs excluded from initial filters'.format(df_original['remove'].sum()))
    # store a copy of the columns and their order for later
    df_cols = df_adjusted.columns
    # concatenate data frames
    df_adjusted = pd.concat([ \
      df_adjusted, \
      df_original.loc[df_original['remove']]], 
      axis=0, ignore_index=True, sort=True)
    # pd.concat reindexes the order of the columns, 
    # so just order it back to what it used to be
    df_adjusted = df_adjusted.reindex(df_cols, axis=1)

  # sort by ID, and reset index
  df_adjusted = df_adjusted.sort_values(['id'])
  df_adjusted = df_adjusted.reset_index(drop=True)

  # add residual RT (alignment error) column
  df_adjusted['residual'] = np.abs(\
    df_adjusted[config['col_names']['retention_time']] - df_adjusted['muij'])

  # add dart_PEP column - which is pep_new, with the NaNs filled in
  # with the old PEPs.
  df_adjusted['dart_PEP'] = df_adjusted['pep_new']
  df_adjusted['dart_PEP'][pd.isnull(df_adjusted['pep_new'])] = \
    df_adjusted[config['col_names']['pep']][pd.isnull(df_adjusted['pep_new'])]
  # make sure that updated PEP does not exceed 1
  df_adjusted['dart_PEP'][df_adjusted['dart_PEP'] > 1] = 1

  # add q-value (FDR) column
  # rank-sorted, cumulative sum of PEPs is expected number of false positives
  # q-value is just that vector divided by # of observations, to get FDR
  logger.info('Calculating FDR (q-values)')
  
  # q-value, without fixing # of false positives to a discrete number
  #df_adjusted['q-value'] = \
  #  ( \
  #    np.cumsum(df_adjusted['dart_PEP'][np.argsort(df_adjusted['dart_PEP'])]) / \
  #    np.arange(1, df_adjusted.shape[0]+1) \
  #  )[np.argsort(np.argsort(df_adjusted['dart_PEP']))]
  
  # q-value, by fixing # of false positives to a discrete number
  # for now, set all null PEPs to 1. we'll remember the index and set them back to nan later
  null_peps = pd.isnull(df_adjusted['dart_PEP'])
  if null_peps.sum() > 0:
    df_adjusted['dart_PEP'][null_peps] = 1

  # get the index order of sorted PEPs
  pep_order = np.argsort(df_adjusted['dart_PEP'])
  # Take the ceiling of the cumulative sum of the sorted PEPs to get the pessimistic
  # estimate of the number of false positives when selecting at that level.
  # because using ceiling, PSMs with different PEPs but within the same relative interval
  # will get the same "num_fp" value.
  num_fp = np.ceil(np.cumsum(df_adjusted['dart_PEP'][pep_order])).astype(int)
  # count the number of occurrences of num_fp and sum them up to get the sample size for each
  # discrete false positive # threshold
  fp_counts = np.cumsum(num_fp.value_counts().sort_index()).values
  # divide # of false positivies by sample size to get q-value. sorting the index order brings
  # the order of values back to their original form
  df_adjusted['dart_qval'] = (num_fp / fp_counts[num_fp-1]).values[np.argsort(pep_order.values)]

  # set null PEPs and q-values back to nan
  if null_peps.sum() > 0:
    df_adjusted['dart_PEP'][null_peps] = np.nan
    df_adjusted['dart_qval'][null_peps] = np.nan

  # rename 'remove' column - which indicates whether or not the PSM participated in the
  # DART-ID alignment and update
  df_adjusted['participated'] = ~df_adjusted['remove']

  ## Run protein inference (fido)?
  if 'run_pi' in config and config['run_pi']:
    logger.info('Running protein inference with Fido...')

    # build fido options into a dict (parameter_map)
    parameter_map = {
      'gamma': config['pi_gamma'] if 'pi_gamma' in config else None,
      'alpha': config['pi_alpha'] if 'pi_alpha' in config else None,
      'beta':  config['pi_beta']  if 'pi_beta'  in config else None,

      'connected_protein_threshold':  config['pi_connected_protein_thresh'],
      'omit_clean_peptide_name':     ~config['pi_clean_peptide_name'],
      'all_psms':                     config['pi_use_all_psms'],
      'group_proteins':               config['pi_group_proteins'],
      'prune_low_scores':             config['pi_prune_low_scores'],
      'parameter_accuracy':           config['pi_parameter_accuracy'],

      'proteins_column':         config['col_names']['proteins'],
      'protein_delimiter':       config['pi_protein_delimiter'],
      'leading_protein_column':  config['col_names']['leading_protein'],
      'decoy_tag':               config['pi_decoy_tag'],

      'sequence_column':         config['col_names']['sequence'],
      #'error_prob_column':       config['col_names']['pep']
      'error_prob_column':       'dart_PEP',

      # pass in output folder so fido can save some intermediate and output files
      'output': config['output']
    }
    logger.debug('parameter_map for fido:')
    logger.debug(str(parameter_map))

    # run fido subroutine
    df_adjusted = run_internal(df_adjusted, parameter_map)

    logger.info('Fido finished')
    logger.info('FDR for PSM\'s razor protein, from protein inference, placed in \"razor_protein_fdr\" column')

  # print figures?
  if config['print_figures']:
    figures(df_adjusted, config, params)

  # overwrite PEP?
  # if true, then store old PEP in "Spectra PEP" column,
  # and put the dart PEP in "PEP" column.
  # then drop the pep_new and dart_PEP columns
  if config['overwrite_pep']:
    logger.info('Overwriting PEP column with new PEP. Saving old PEP in \"Spectra PEP\" column.')
    df_adjusted['Spectra PEP'] = df_adjusted[config['col_names']['pep']]
    df_adjusted[config['col_names']['pep']] = df_adjusted['dart_PEP']
    df_adjusted = df_adjusted.drop(['pep_new', 'dart_PEP'], axis=1)

  # tell the user whether or not to expect diagnostic columns
  if config['add_diagnostic_cols']:
    logger.info('Adding diagnostic columns to output')

  # if no output is specified, throw an error
  if config['save_combined_output'] == False and config['save_separate_output'] == False:
    raise ConfigFileError('No output format specified. Either set \"save_combined_output\" to true, or set \"save_separate_output\" to true.')

  # write to file
  if config['save_combined_output']:
    # if combining input files, then write to one combined file
    out_path = os.path.join(config['output'], config['combined_output_name'])
    logger.info('Combining input file(s) and writing adjusted data file to {} ...'.format(out_path))
    write_output(df_adjusted, out_path, config)
  
  if config['save_separate_output']:
    # if keeping input files separate, then use 'input_id' to retain the
    # order in which the input files were passed in
    logger.info('Saving output to separate files...')
    for i, f in enumerate(config['input']):

      # get output extension
      # default to the same extension as the input
      # if one in the config file exists, use that instead
      out_ext = os.path.splitext(os.path.basename(f))[1]
      if config['output_ext'] is not None:
        out_ext = config['output_ext']

      # contruct output path based on which input file it was
      out_path = os.path.join(config['output'], 
        os.path.splitext(os.path.basename(f))[0] + \
        config['output_suffix'] + '_' + str(i) + out_ext)

      # if saving back to the original input folder,
      # then base the output file name on the input file name instead.
      # no need to number these.
      if config['save_in_input_folder']:
        out_path = os.path.join(os.path.dirname(f),
          os.path.splitext(os.path.basename(f))[0] + \
          config['output_suffix'] + out_ext)

      logger.info('Saving input file {} to {}'.format(i, out_path))
      df_a = df_adjusted.loc[df_adjusted['input_id'] == i]
      # save to file
      # other data formats might have a different separator, or have an index column
      write_output(df_a, out_path, config)

  logger.info('Done! Process took {:.3f} seconds'.format(time.time() - start))

if __name__ == '__main__':
  main()







# coding: utf-8
#!/usr/bin/env python3
#
# Copyright (c) 2008, Bjoern B. Brandenburg <bbb [at] cs.unc.edu>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS  PROVIDED BY THE COPYRIGHT HOLDERS  AND CONTRIBUTORS "AS IS"
# AND ANY  EXPRESS OR  IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED  TO, THE
# IMPLIED WARRANTIES  OF MERCHANTABILITY AND  FITNESS FOR A  PARTICULAR PURPOSE
# ARE  DISCLAIMED. IN NO  EVENT SHALL  THE COPYRIGHT  OWNER OR  CONTRIBUTORS BE
# LIABLE  FOR   ANY  DIRECT,  INDIRECT,  INCIDENTAL,   SPECIAL,  EXEMPLARY,  OR
# CONSEQUENTIAL  DAMAGES  (INCLUDING,  BUT   NOT  LIMITED  TO,  PROCUREMENT  OF
# SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA,  OR PROFITS;  OR BUSINESS
# INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY  THEORY OF  LIABILITY, WHETHER  IN
# CONTRACT,  STRICT  LIABILITY, OR  TORT  (INCLUDING  NEGLIGENCE OR  OTHERWISE)
# ARISING IN ANY  WAY OUT OF THE USE  OF THIS SOFTWARE, EVEN IF  ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# Edited, 2018, Albert Chen @ Northeastern

"""An implementation of Binomial Heaps.

From Wikipedia:
  A binomial heap is a heap similar to a binary heap but also supporting the
  operation of merging two heaps quickly. This is achieved by using a special
  tree structure.

  All of the following operations work in O(log n) time on a binomial heap with
  n elements:
    - Insert a new element to the heap
    - Find the element with minimum key
    - Delete the element with minimum key from the heap
    - Decrease key of a given element
    - Delete given element from the heap
    - Merge two given heaps to one heap

  More details: http://en.wikipedia.org/wiki/Binomial_heap

This implementation is based on the description in CLRS.
"""

class ItemRef(object):
  """Reference to an item in the heap. Used for decreasing keys and deletion.
  Do not use this class directly; only use instances returned by
  BinomialHeap.insert()!

  You should only use ItemRef.delete() and ItemRef.decrease(new_priority).
  """
  def __init__(self, node, get_heap):
    self.ref      = node
    self.get_heap = get_heap
    self.in_tree  = True

  def __str__(self):
    if self.in_tree: return "<BinomialHeap Reference to '%s'>" % str(self.ref.val)
    else:            return "<stale BinomialHeap Reference>"

  def decrease(self, new_key):
    "Update the priority of the referenced item to a lower value."
    assert self.in_tree
    assert self.ref.ref == self
    self.ref.decrease(new_key)

  def delete(self):
    """Remove the referenced item from the heap.
    """
    self.decrease(self)
    v = self.get_heap().extract_min()
    assert not self.in_tree
    assert v is self.ref.val

  def in_heap(self, heap):
    """Returns True if the referenced item is part of the BinomialHeap 'heap';
    False otherwise.
    """
    return self.in_tree and self.get_heap() == heap

  def __lt__(self, other):
    "Behaves like negative infinity: always True."
    return True

  def __gt__(self, other):
    "Behaves like negative infinity: always False."
    return False


class BinomialHeap(object):
  """Usage:
  > H1 = BinomialHeap()
  > H1.insert(40, "fast.")
  > H1.insert(10, "Merging")
  > H2 = BinomialHeap([(30, "quite"), (20, "is")])
  > H1 += H2
  > for x in H1:
  >   print x,
   =>  "Merging is quite fast."

  """

  class Node(object):
    "Internal node of the heap. Don't use directly."
    def __init__(self, get_heap, key, val=None):
      self.degree = 0
      self.parent = None
      self.next   = None
      self.child  = None
      self.key    = key
      self.ref    = ItemRef(self, get_heap)
      if val == None: val = key
      self.val    = val

    def __str__(self):
      k = lambda x: str(x.key) if x else 'NIL'
      return '(%s, c:%s, n:%s)' % (k(self), k(self.child), k(self.next))

    def link(self, other):
      "Makes other a subtree of self."
      other.parent  = self
      other.next    = self.child
      self.child    = other
      self.degree  += 1

    def decrease(self, new_key):
      node = self
      assert new_key < node.key
      node.key = new_key
      cur    = node
      parent = cur.parent
      while parent and cur.key < parent.key:
        # need to bubble up
        # swap refs
        parent.ref.ref, cur.ref.ref = cur, parent
        parent.ref, cur.ref         = cur.ref, parent.ref
        # now swap keys and payload
        parent.key, cur.key         = cur.key, parent.key
        parent.val, cur.val         = cur.val, parent.val
        # step up
        cur    = parent
        parent = cur.parent

    @staticmethod
    def roots_merge(h1, h2):
      """Merge two lists of heap roots, sorted by degree.
      Returns the new head.
      """
      if not h1: return h2
      if not h2: return h1
      if h1.degree < h2.degree:
        h  = h1
        h1 = h.next
      else:
        h  = h2
        h2 = h2.next
      p = h
      while h2 and h1:
        if h1.degree < h2.degree:
          p.next = h1
          h1 = h1.next
        else:
          p.next = h2
          h2 = h2.next
        p = p.next
      if h2:
        p.next = h2
      else:
        p.next = h1
      return h

    @staticmethod
    def roots_reverse(h):
      """Reverse the heap root list.
      Returns the new head. Also clears parent references.
      """
      if not h: return None
      tail = None
      next = h
      h.parent = None
      while h.next:
        next = h.next
        h.next = tail
        tail   = h
        h = next
        h.parent = None
      h.next = tail
      return h

  class __Ref(object):
    def __init__(self, h):
      self.heap = h
      self.ref  = None
    def get_heap_ref(self):
      if not self.ref: return self
      else:
        # compact
        self.ref  = self.ref.get_heap_ref()
        return self.ref
    def get_heap(self):
      return self.get_heap_ref().heap

  def __init__(self, lst=[]):
    """Populate a new heap with the (key, value) pairs in 'lst'.
    If the elements of lst are not subscriptable, then they are treated as
    opaque elements and inserted into the heap themselves.
    """
    self.head = None
    self.size = 0
    self.ref  = BinomialHeap.__Ref(self)
    for x in lst:
      try:
        self.insert(x[0], x[1])
      except TypeError:
        self.insert(x)

  def insert(self, key, value=None):
    """Insert 'value' in to the heap with priority 'key'. If 'value' is omitted,
    then 'key' is used as the value.
    Returns a reference (of type ItemRef) to the internal node in the tree.
    Use this reference to delete the key or to change its priority.
    """
    n = BinomialHeap.Node(self.ref.get_heap, key, value)
    self.__union(n)
    self.size += 1
    return n.ref

  def union(self, other):
    """Merge 'other' into 'self'. Returns None.
    Note: This is a destructive operation; 'other' is an empty heap afterwards.
    """
    self.size = self.size + other.size
    h2        = other.head
    self.__union(h2)
    other.ref.ref = self.ref
    other.__init__()

  def min(self):
    """Returns the value with the minimum key (= highest priority) in the heap
    without removing it, or None if the heap is empty.
    """
    pos = self.__min()
    return pos[0].val if pos else None

  def extract_min(self):
    """Returns the value with the minimum key (= highest priority) in the heap
    AND removes it from the heap, or None if the heap is empty.
    """
    # find mininum
    pos = self.__min()
    if not pos: return None
    else:
      (x, prev) = pos
      # remove from list
      if prev:
          prev.next = x.next
      else:
          self.head = x.next
      kids = BinomialHeap.Node.roots_reverse(x.child)
      self.__union(kids)
      x.ref.in_tree = False
      self.size -= 1
      return x.val

  def __bool__(self):
    """True if the heap is not empty; False otherwise."""
    return self.head != None

  def __iter__(self):
    """Returns a _destructive_ iterator over the values in the heap.
    This violates the iterator protocol slightly, but is very useful.
    """
    return self

  def __len__(self):
    """Returns the number of items in this heap."""
    return self.size

  def __setitem__(self, key, value):
    """Insert.
    H[key] = value  is equivalent to  H.insert(key, value)
    """
    self.insert(key, value)

  def __iadd__(self, other):
    """Merge.
    a += b  is equivalent to  a.union(b).
    """
    self.union(other)
    return self

  def __next__(self):
    """Returns the value with the minimum key (= highest priority) in the heap
    AND removes it from the heap; raises StopIteration if the heap is empty.
    """
    if self.head:
        return self.extract_min()
    else:
        raise StopIteration

  def __contains__(self, ref):
    """Test whether a given reference 'ref' (of ItemRef) is in this heap.
    """
    if type(ref) != ItemRef:
        raise TypeError("Expected an ItemRef")
    else:
        return ref.in_heap(self)

  def __min(self):
    if not self.head: return None
    min  = self.head
    min_prev = None
    prev = min
    cur  = min.next
    while cur:
      if cur.key < min.key:
        min = cur
        min_prev = prev
      prev = cur
      cur  = cur.next
    return (min, min_prev)

  def __union(self, h2):
    if not h2: return # nothing to do
    h1 = self.head
    if not h1:
      self.head = h2
      return
    h1 = BinomialHeap.Node.roots_merge(h1, h2)
    prev = None
    x    = h1
    next = x.next
    while next:
      if x.degree != next.degree or \
         (next.next and next.next.degree == x.degree):
        prev = x
        x    = next
      elif x.key <= next.key:
        # x becomes the root of next
        x.next = next.next
        x.link(next)
      else:
        # next becomes the root of x
        if not prev:
          # update the "master" head
          h1 = next
        else:
          # just update previous link
          prev.next = next
        next.link(x)
        # x is not toplevel anymore, update ref by advancing
        x = next
      next = x.next
    self.head = h1

def heap(lst=[]):
    """Create a new heap. lst should be a sequence of (key, value) pairs.
    Shortcut for BinomialHeap(lst)
    """
    return BinomialHeap(lst)

if __name__ == "__main__":
  tokens1 = [(24, 'all'), (16, 'star'), (9, 'true.\nSinging'), (7, 'clear'),
             (25, 'praises'), (13, 'to'), (5, 'Heel'),
             (6, 'voices\nRinging'), (26, 'thine.'), (21, 'shine\nCarolina'),
             (117, 'Rah,'), (102, 'Tar'), (108, 'bred\nAnd'), (125, 'Rah!'),
             (107, 'Heel'), (118, 'Rah,'), (111, "die\nI'm"),
             (115, 'dead.\nSo'), (120, 'Rah,'), (121, "Car'lina-lina\nRah,"),
             (109, 'when'), (105, 'a'), (123, "Car'lina-lina\nRah!"),
             (110, 'I'), (114, 'Heel'), (101, 'a'), (106, 'Tar'),
             (18, 'all\nClear'), (14, 'the')]
  tokens2 = [(113, 'Tar'), (124, 'Rah!'), (112, 'a'), (103, 'Heel'),
             (104, "born\nI'm"), (122, 'Rah,'), (119, "Car'lina-lina\nRah,"),
             (2, 'sound'), (20, 'radiance'), (12, 'N-C-U.\nHail'),
             (10, "Carolina's"), (3, 'of'), (17, 'of'),
             (23, 'gem.\nReceive'), (19, 'its'), (0, '\nHark'),
             (22, 'priceless'), (4, 'Tar'), (1, 'the'), (8, 'and'),
             (15, 'brightest'), (11, 'praises.\nShouting'),
             (100, "\nI'm"), (116, "it's")]
  h1 = heap(tokens1)
  h2 = heap(tokens2)
  h3 = heap()
  line = "\n==================================="
  h3[90]  = line
  h3[-2]  = line
  h3[200] = line
  h3[201] = '\n\n'
  t1ref = h3.insert(1000, "\nUNC Alma Mater:")
  t2ref = h3.insert(120, "\nUNC Fight Song:")
  bad   = [h3.insert(666, "Dook"),
           h3.insert(666, "Go Devils!"),
           h3.insert(666, "Blue Devils") ]

  ref = bad[0]
  print("%s: \n\tin h1: %s\n\tin h2: %s\n\tin h3: %s" % \
    (str(ref), ref in h1, ref in h2, ref in h3))

  print("Merging h3 into h2...")
  h2 += h3

  print("%s: \n\tin h1: %s\n\tin h2: %s\n\tin h3: %s" % \
    (str(ref), ref in h1, ref in h2, ref in h3))

  print("Merging h2 into h1...")
  h1 += h2

  print("%s: \n\tin h1: %s\n\tin h2: %s\n\tin h3: %s" % \
    (str(ref), ref in h1, ref in h2, ref in h3))

  t1ref.decrease(-1)
  t2ref.decrease(99)

  for ref in bad:
    ref.delete()
  for x in h1:
    print(x, end=' ')




#!/usr/bin/env python3
# coding: utf-8

import logging
import matplotlib
matplotlib.use('PS')
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd

from dart_id.helper import create_fig_folder
from scipy.stats import norm, lognorm, laplace

logger = logging.getLogger('root')

def gen(df, config, params, output_path):
  df = df[-(df['remove'])]

  figures_path = create_fig_folder(output_path, 'figures')
  fig_names = []

  exp_names = np.sort(df[config['col_names']['raw_file']].unique())
  num_experiments = len(exp_names)
  # ceil PEP to 1
  df[config['col_names']['pep']].loc[df[config['col_names']['pep']] > 1] = 1
  # split PEP into 10 bins, for coloring points later
  #pep_col_code = pd.cut(df[config['col_names']['pep']], 10)
  pep_col_code = pd.cut(df[config['col_names']['pep']], np.linspace(0, 1, 11))

  # generate figures for each experiment
  for exp in range(0, num_experiments):
    logger.info('Generating Summary for Experiment {} | {}'.format(exp+1, exp_names[exp]))

    exp_params = params['exp'].iloc[exp]
    #exp_indices = params['pair']['muij_to_exp'] == exp
    exp_inds = (df['exp_id'] == exp) & (~pd.isnull(df['pep_new']))

    # dont plot if there aren't any points
    if np.sum(exp_inds) == 0: continue

    predicted = df['muij'][exp_inds].values
    predicted_sd = df['sigmaij'][exp_inds].values
    mus = df['mu'][exp_inds].values

    # observed values
    observed = df[config['col_names']['retention_time']][exp_inds].values
    obs_peps = df[config['col_names']['pep']][exp_inds].values
    obs_code = pep_col_code[exp_inds].values
    residual = observed - predicted

    # plot the 2-segment linear fit of mus to observed RTs
    plt.subplot(121)
    plt.scatter(mus, observed, s=1, color='black')
    plt.plot([0, exp_params['split_point']],
             [exp_params['beta_0'], (exp_params['split_point'] * exp_params['beta_1']) + exp_params['beta_0']],
            color='red')
    plt.plot([exp_params['split_point'], 300], 
             [(exp_params['split_point'] * exp_params['beta_1']) + exp_params['beta_0'], (exp_params['split_point'] * exp_params['beta_1']) + ((300-exp_params['split_point']) * exp_params['beta_2']) + exp_params['beta_0']],
            color='green')
    plt.plot(np.repeat(exp_params['split_point'], 2), [-100, 300], color='blue', linestyle='dashed')
    plt.axis([0, mus.max() + 10, exp_params['beta_0']-10, observed.max() + 10])
    plt.title('Experiment {} - {}'.format(exp, exp_names[exp]))
    plt.xlabel('Reference RT (min)')
    plt.ylabel('Observed RT (min)')


    # plot residuals, quantiles of residuals, and color points by PEP
    plt.subplot(122)
    plt.scatter(predicted, residual, s=4, c=pep_col_code.cat.codes.values[exp_inds], alpha=0.5)
    plt.plot([0, exp_params["split_point"]], [0, 0], color="red")
    plt.plot([exp_params["split_point"], 300], [0, 0], color="green")
    plt.plot(np.repeat(exp_params["split_point"], 2), [-100, 300], color="blue", linestyle="dashed")
    
    # confidence intervals, 2.5% and 97.5%
    conf_x = predicted[np.argsort(predicted)]
    conf_2p5 = laplace.ppf(0.025, loc=0, scale=predicted_sd)[np.argsort(predicted)]
    conf_97p5 = laplace.ppf(0.975, loc=0, scale=predicted_sd)[np.argsort(predicted)]
    
    plt.plot(conf_x, conf_2p5, color="red")
    plt.plot(conf_x, conf_97p5, color="red")
    plt.axis([predicted.min()-5, predicted.max()+5, residual.min()-5, residual.max()+5])
    
    plt.ylim(np.min(conf_2p5) - 0.1, np.max(conf_97p5) + 0.1)
    plt.xlim(conf_x[0], conf_x[-1])
    
    cbar = plt.colorbar()
    cbar.set_label('Spectral PEP (Error Probability)')
    cbar.ax.set_yticklabels(pep_col_code.cat.categories.values)
    plt.xlabel("Inferred RT (min)")
    plt.ylabel("Residual RT (min)")

    # add some space between subplots
    plt.subplots_adjust(hspace=0.3, wspace=0.35, bottom=0.2, right=0.85)

    # finalize and save figure
    fig = plt.gcf()
    fig.set_size_inches(7, 3.5)
    _fname = 'alignment_{}_{}.png'.format(str(exp), exp_names[exp])
    fname = os.path.join(figures_path, _fname)
    logger.info('Saving figure to {} ...'.format(fname))
    fig.savefig(fname, dpi=160)
    fig_names.append(os.path.join('figures', _fname))
    
    plt.close()
    fig.clf()

  return fig_names




#!/usr/bin/env python3
# coding: utf-8
# 
# Downloaded from https://noble.gs.washington.edu/proj/fido/
# Fido is described in this paper: http://dx.doi.org/10.1021/pr100594k
# Modified 2018 by Albert Chen, Northeastern

import argparse
import getopt
import io
import itertools
import logging
import networkx as nx
import os
import pandas as pd
import random
import sys

from dart_id.fido.Utilities import *
from dart_id.fido.GraphUtilities import *
from dart_id.helper import *
from pprint import pprint
from time import time

logger = logging.getLogger('root')

class potential_table:
  def __init__(self, domain):
    self.potentials_on_frozensets = {}
    self.domain = domain
  def tuple_to_domain_frozenset(self, t):
    return frozenset( [(var,value) for var,value in t if var in self.domain] )
  def __getitem__(self, item):
    return self.potentials_on_frozensets[self.tuple_to_domain_frozenset(item)]
  def __setitem__(self, item, value):
    self.potentials_on_frozensets[self.tuple_to_domain_frozenset(item)] = value
  def __contains__(self, item):
    return self.tuple_to_domain_frozenset(item) in self.potentials_on_frozensets

class distribution(potential_table):
  def __init__(self, domain_vars_to_outcomes):
    logger.info('\t\tmaking distribution on {} variables'.format(len(domain_vars_to_outcomes)))
    potential_table.__init__(self, list(domain_vars_to_outcomes) )
    self.domain_vars_to_outcomes = domain_vars_to_outcomes
  def get_all_possible(self):
    outcomes = list( itertools.product(*[ single_outcome for var_name,single_outcome in list(self.domain_vars_to_outcomes.items())]) )
    var_outcome_lists = [ list(zip(list(self.domain_vars_to_outcomes), o)) for o in outcomes ]
    return var_outcome_lists
  def display(self):
    for o in self.get_all_possible():
      logger.info('{} --> {}'.format(o, self[o]))
  def marginalized_out(self, var_set):
    remaining = set(self.domain_vars_to_outcomes).difference(var_set)
    marginal_distribution = distribution( dict([ (v, self.domain_vars_to_outcomes[v]) for v in remaining ]) )
    for rc in marginal_distribution.get_all_possible():
      marginal_distribution[rc] = 0.0
    total = 0.0
    for c in self.get_all_possible():
      value_at_c = self[c]
      marginal_distribution[c] += value_at_c
      total += value_at_c
    for rc in marginal_distribution.get_all_possible():
      marginal_distribution[rc] /= total
    return marginal_distribution
  def __mul__(self, other_distribution):
    return distribution.multiply([self, other_distribution])
  def __div__(self, other_distribution):
    ### other_distribution should have a domain that is a subset
    ### of the current distribution
    new_distribution = distribution(self.domain_vars_to_outcomes)
    for o in self.get_all_possible():
    ### note: this will not work if other_distribution contains
    ### any 0 values; but in those cases, any value will work for
    ### belief propagation because it will be multiplied by zero
      if abs(other_distribution[o]) < 1e-5:
        #new_distribution[o] = 1
        new_distribution[o] = 0
      else:
        new_distribution[o] = self[o] / other_distribution[o]
    return new_distribution
  def distribution_constant(self):
    tot = 0.0
    for o in self.all_outcomes():
      tot += self[o]
    return tot
  @staticmethod
  def multiply(distribution_list):
    new_domain_vars_to_outcomes = {}
    for d in distribution_list:
      new_domain_vars_to_outcomes.update(d.domain_vars_to_outcomes)
    joint_product_distribution = distribution(new_domain_vars_to_outcomes)
    for c in joint_product_distribution.get_all_possible():
      prod = 1.0
      for d in distribution_list:
        prod *= d[c]
      joint_product_distribution[c] = prod
    return joint_product_distribution

class identifier(hashable_dict):
  def __str__(self):
    return 'identifier(' + str(self.my_dict) + ')'

def fs_node_ids(fs):
  return [n['id'] for n in fs]

class node:
  def __init__(self, identity, outcomes = (True, False)):
    self.id = identity
    self.outcomes = outcomes
    self.parameter_names_to_nodes = None
  def __str__(self):
    return 'node(' + str(self.id) + ', param_map = ' + str(self.parameter_names_to_nodes) + ')'
  def __repr__(self):
    return self.__str__()
  def potential(self, **Kwargs):
    raise Exception('node::potential is pure virtual')
  def wrapped_potential(self, nodes_to_values):
    ### Use the nodes_to_values dict the parameter_names_to_nodes
    ### multi_dict to make a dict of parameter_names_to_values
    #logger.info('computing potential for node {}'.format(self.id))
    #logger.info('nodes to values {}'.format(nodes_to_values))
    #logger.info('names to nodes {}'.format(self.parameter_names_to_nodes))
    parameter_names_to_values = {}
    for parameter, nodes in list(self.parameter_names_to_nodes.items()):
      parameter_node_values = [ nodes_to_values[n] for n in nodes ]
      if len(parameter_node_values) > 1:
        parameter_names_to_values[ parameter ] = parameter_node_values
      else:
        parameter_names_to_values[ parameter ] = parameter_node_values[0]

    #print(parameter_names_to_values)
    return self.potential(**parameter_names_to_values)
  def __hash__(self):
    return hash(self.id)
  def __eq__(self, rhs):
    return isinstance(rhs, node) and self.id == rhs.id
  def __ne__(self, rhs):
    return not self == rhs

class iid_probability_node(node):
  gamma = 0.5
  def __init__(self, id):
    node.__init__(self, id, (True, False))
  def potential(self, my_value, **Kwargs):
    if my_value == True:
      return iid_probability_node.gamma
    return 1-iid_probability_node.gamma

class noisy_or_node(node):
  alpha = 0.25
  beta = 0.01
  def __init__(self, id):
    node.__init__(self, id, (True, False))
  def potential(self, input, my_value, **Kwargs):
    ### when there is a single input, you should convert it to a
    ### list so that you can measure its cardinality
    if not isinstance(input, list):
      input = [input]
    count = len([ i for i in input if i == True ])
    false_prob = (1-noisy_or_node.beta)* ( (1-noisy_or_node.alpha)**count )
    if my_value == False:
      return false_prob
    return 1-false_prob

class predecessor_table_node(node):
  def __init__(self, id, likelihood_table):
    node.__init__(self, id, (True,))
    self.likelihood_table = likelihood_table
  def potential(self, input, my_value, **Kwargs):
    return self.likelihood_table[my_value][input]

def bayesian_network_union(bn1, bn2):
  if bn1 == None: return bn2
  if bn2 == None: return bn1
  for i in bn2:
    bn1.add_node(bn2.get_node_from_id(i), **bn2.node[i])
    for j in bn2[i]:
      bn1.add_edge(bn2.get_node_from_id(i),bn2.get_node_from_id(j),**bn2[i][j])
  return bn1

def all_bayesian_network_unions(graph_lst):
  ### do not set as Graph or DiGraph so that it works with any
  ### homogeneous list
  result = None
  for g in graph_lst:
    result = bayesian_network_union(result, g)
  return result

### bayesian_network should only accept 'node' types (but can be
### indexed by identifier class)
class bayesian_network(nx.DiGraph):
  def __init__(self, **Kwargs):
    nx.DiGraph.__init__(self, **Kwargs)
    self.id_to_node_map = {}

  def __str__(self):
    sio = io.StringIO()
    for n_id in self:
      sio.write(n_id['id'] + ': ')
      targets = []
      for tgt in self[n_id]:
        targets.append(tgt['id'])

      for t in targets:
        sio.write(t)
    return sio.getvalue()
    #return 'a'

  def get_potential_distribution_for_node(self, n_id):
    n_node = self.get_node_from_id(n_id)
    pred = self.predecessors(n_id)
    for m_id in pred:
      m_node = self.get_node_from_id(m_id)
          

  def subgraph(self, ibunch):
    ### using the self.__class__() constructor will work properly
    ### when inherited from; for instance, when called from a
    ### fido_network, it should allocate sg as a new fido_network
    
    sg = self.__class__()

    identifier_set = set([i for i in ibunch])
    for i in ibunch:
      sg.add_node(self.id_to_node_map[i], **self.node[i])
      for tgt in self[i]:
        if tgt not in identifier_set: continue
        sg.add_edge(self.id_to_node_map[i], self.id_to_node_map[tgt], **self[i][tgt])

    return sg

  def get_node_from_id(self, ident):
    return self.id_to_node_map[ident]

  ### method to maintain the id_to_node_map after a change to the
  ### graph
  def update_id_to_node_map_after_change(self, all_changed_nodes):
    for n in all_changed_nodes:
      if n.id in self and n.id not in self.id_to_node_map:
        self.id_to_node_map[n.id] = n
      elif n.id not in self and n.id in self.id_to_node_map:
        del self.id_to_node_map[n.id]

  ### overridden functions for addition and removal: these functions
  ### maintain the graph and the id : node map
  def add_node(self, node_instance, **Kwargs):
    assert isinstance(node_instance, node)
    nx.DiGraph.add_node(self, node_instance.id, **Kwargs)
    self.update_id_to_node_map_after_change((node_instance,))
  def remove_node(self, node_instance, **Kwargs):
    assert isinstance(node_instance, node)
    nx.DiGraph.remove_node(self, node_instance.id, **Kwargs)
    self.update_id_to_node_map_after_change((node_instance,))
  def add_edge(self, node_a, node_b, **Kwargs):
    assert isinstance(node_a, node)
    assert isinstance(node_b, node)
    nx.DiGraph.add_edge(self, node_a.id, node_b.id, **Kwargs)
    self.update_id_to_node_map_after_change((node_a, node_b))
  def remove_edge(self, node_a, node_b, **Kwargs):
    assert isinstance(node_a, node)
    assert isinstance(node_b, node)
    nx.DiGraph.remove_edge(self, node_a.id, node_b.id, **Kwargs)
    self.update_id_to_node_map_after_change((node_a, node_b))
  def add_nodes_from(self, node_lst, **Kwargs):
    for n in node_lst:
      assert isinstance(n, node)
    id_lst = [n.id for n in node_lst]
    nx.DiGraph.add_nodes_from(self, id_lst, **Kwargs)
    self.update_id_to_node_map_after_change(node_lst)
  def remove_nodes_from(self, node_lst, **Kwargs):
    for n in node_lst:
      assert isinstance(n, node)
    id_lst = [n.id for n in node_lst]
    nx.DiGraph.remove_nodes_from(self, id_lst, **Kwargs)
  def add_edges_from(self, edge_lst, **Kwargs):
    for e in edge_lst:
      n_a, n_b = e
      assert isinstance(n_a, node)
      assert isinstance(n_b, node)
    id_lst = [(n_a.id,n_b.id) for (n_a,n_b) in node_lst]
    nx.DiGraph.add_edges_from(self, id_lst, **Kwargs)
    self.update_id_to_node_map_after_change([n_a for n_a,n_b in edge_lst])
    self.update_id_to_node_map_after_change([n_b for n_a,n_b in edge_lst])
  def remove_edges_from(self, edge_lst, **Kwargs):
    for e in edge_lst:
      n_a, n_b = e
      assert isinstance(n_a, node)
      assert isinstance(n_b, node)
    id_lst = [(n_a.id,n_b.id) for (n_a,n_b) in node_lst]
    nx.DiGraph.remove_edges_from(self, id_lst, **Kwargs)
    self.update_id_to_node_map_after_change([n_a for n_a,n_b in edge_lst])
    self.update_id_to_node_map_after_change([n_b for n_a,n_b in edge_lst])

  def get_parameter_names_to_nodes_map(self, i):
    input_map = multi_dict()
    pred = self.predecessors(i)
    for p in pred:
      input_map.add( self[p][i]['label'], p )
    # each node depends on its own value
    input_map.add('my_value', i)
    return input_map
  def init_parameter_names_to_nodes_maps(self):
    for i in self:
      n = self.get_node_from_id(i)
      n.parameter_names_to_nodes = self.get_parameter_names_to_nodes_map(i)

class fido_network(bayesian_network):
  def __init__(self, fido_connected_protein_threshold = 15, \
    omit_clean_peptide_name = False, \
    proteins_column = 'Proteins', protein_delimiter = ';', \
    leading_protein_column = 'Leading razor protein', decoy_tag = 'REV__', 
    sequence_column = 'Sequence', error_prob_column = 'PEP', **Kwargs):

    bayesian_network.__init__(self, **Kwargs)
    self.fido_connected_protein_threshold = fido_connected_protein_threshold
    self.clean_peptide_id = ~omit_clean_peptide_name

    self.proteins_column = proteins_column
    self.protein_delimiter = protein_delimiter
    self.leading_protein_column = leading_protein_column
    self.decoy_tag = decoy_tag

    self.sequence_column = sequence_column
    self.error_prob_column = error_prob_column

  @staticmethod
  def likelihood_term(connected_fido_network, pep, pep_state, protein_config):
    ### computes the likelihood term from one peptide (proteins
    ### --> peptide and peptide --> spectrum) in a given state
    pep_node = connected_fido_network.get_node_from_id(pep)

    spectrum_id = list(connected_fido_network.successors(pep))[0]
    spectrum_node = connected_fido_network.get_node_from_id(spectrum_id)
    prop_prob_spectrum_given_peptide_states = spectrum_node.likelihood_table[True]
    
    protein_config[pep] = pep_state
    prob_peptide_state_given_proteins = pep_node.wrapped_potential(protein_config)
    prob_at_pep_state = prob_peptide_state_given_proteins * prop_prob_spectrum_given_peptide_states[pep_state]
    del protein_config[pep]

    return prob_at_pep_state

  @staticmethod
  def likelihood_protein_configuration(connected_fido_network, protein_config):
    ### go through the peptide terms and get sum_e Pr(E=e | R=r) and
    ### Pr(D_i | E=e)
    peps = [ pep for pep in connected_fido_network if pep['type'] == 'peptide' ]
    likelihood = 1.0
    for pep in peps:
      likelihood *= sum([ fido_network.likelihood_term(connected_fido_network, pep, pep_state, protein_config) for pep_state in (True,False) ])
      #likelihood *= sum([ prob_peptide_states[i] * prop_prob_spectrum_given_peptide_states[i] for i in [False, True]])
    return likelihood

  @staticmethod
  def prob_protein_configuration(connected_fido_network, protein_config):
    n = len(protein_config)
    r = len([ (prot,state) for prot,state in list(protein_config.items()) if state == True ])
    return pow(iid_probability_node.gamma, r) * pow(1-iid_probability_node.gamma, n-r)


  def FidoMarginalization_Inference(self):
    ### sum over the power set of proteins (for each protein
    ### configuration, compute the likelihood for the peptide
    ### sets)

    all_posteriors = {}

    subgraphs = nx.weakly_connected_component_subgraphs(self, copy=False)
    for sg_number, sg in enumerate(subgraphs):

      #if log_connected_naive_complexity(sg) > 13:
      #logger.info('fido is marginalizing subgraph with {}'.format(log_connected_naive_complexity(sg)))

      prot_nodes = [ p for p in sg if p['type'] == 'protein' ]
      prot_outcomes = itertools.product(*[(True, False) for p in prot_nodes])

      subgraph_posteriors = dict.fromkeys(prot_nodes, 0.0)
      total_prop_sum = 0.0
      for outcome in prot_outcomes:
        outcome_map = dict(list(zip(prot_nodes,outcome)))
        #print(outcome_map)
        likelihood = fido_network.likelihood_protein_configuration(sg, outcome_map)
        #print(likelihood)
        prior = fido_network.prob_protein_configuration(sg, outcome_map)
        prop_posterior = likelihood * prior
        total_prop_sum += prop_posterior
        for prot,state in list(outcome_map.items()):
          if state == True: subgraph_posteriors[prot] += prop_posterior

      for prot in subgraph_posteriors:
        subgraph_posteriors[prot] /= total_prop_sum
      all_posteriors.update(subgraph_posteriors)


    return all_posteriors

  @staticmethod
  def prune_low_scoring_peptides(dg, score_threshold = 5e-2):
    # go through the nodes with peptide type and call prune_node if
    # the probability is below the threshold
    to_prune_list = []
    for n in dg:
      if n['type'] == 'peptide':
        spectrum_id = list(dg.successors(n))[0]
        spectrum_node = dg.get_node_from_id(spectrum_id)
        if spectrum_node.likelihood_table[True][True] < score_threshold:
          to_prune_list.append(n)
    for pep in to_prune_list:
      fido_network.prune_peptide(dg, pep)
    return dg

  @staticmethod
  def prune_peptide(dg, pep):
    #logger.info('pruning peptide {}'.format(pep))
    associated_proteins = dg.predecessors(pep)
    pep_node = dg.get_node_from_id(pep)
    spect_id = list(dg.successors(pep))[0]
    spect_node = dg.get_node_from_id(spect_id)
    for prot_number, prot in enumerate(associated_proteins):
      prot_node = dg.get_node_from_id(prot)

      new_pep_id_dict = pep.my_dict.copy()
      new_pep_id_dict['copy'] = prot_number
      new_pep_id = identifier(new_pep_id_dict)
      cloned_pep_node = noisy_or_node(new_pep_id)

      new_spect_id_dict = spect_id.my_dict.copy()
      new_spect_id_dict['copy'] = prot_number
      new_spect_id = identifier(new_spect_id_dict)
      cloned_spect_node = predecessor_table_node(new_spect_id, spect_node.likelihood_table)

      dg.add_edge(prot_node, cloned_pep_node, **dg[prot][pep])
      dg.add_edge(cloned_pep_node, cloned_spect_node, **dg[pep][spect_id])
    dg.remove_node(pep_node)
    dg.remove_node(spect_node)

  @staticmethod
  def connected_lowest_prune(connected_dg):
    pep_nodes = [ n for n in connected_dg if n['type'] == 'peptide' and len(list(connected_dg.predecessors(n))) > 1 ]
    pep_present_likelihoods = [ connected_dg.get_node_from_id(list(connected_dg.successors(pep))[0]).likelihood_table[True][True] for pep in pep_nodes ]

    #pep_score, pep_to_prune = np.min(list(zip(pep_present_likelihoods, pep_nodes)))
    pep_to_prune = pep_nodes[np.argmin(pep_present_likelihoods)]

    fido_network.prune_peptide(connected_dg, pep_to_prune)

  @staticmethod
  def dynamic_pruned(dg, connected_protein_threshold, **Kwargs):
    connected_protein_threshold = float(connected_protein_threshold)
    dg_lst = nx.weakly_connected_component_subgraphs(dg, copy=False)
    pruned_dg_lst = fido_network.dynamic_pruned_helper(dg_lst, connected_protein_threshold)
    return all_bayesian_network_unions(pruned_dg_lst)
        
  @staticmethod
  def dynamic_pruned_helper(dg_lst, connected_protein_threshold):
    ### may be slow-- linear time per iteration and possibly many
    ### iterations (--> quadratic in number of peptides in a connected component)
    result_lst = []
    for dg in dg_lst:
      if log_connected_naive_complexity(dg) < connected_protein_threshold:
        result_lst.append(dg)
      else:
        #logger.info('pruning connected graph with {} proteins'.format(log_connected_naive_complexity(dg)))
        while True:
          # prune the lowest-scoring peptide until the graph is split
          fido_network.connected_lowest_prune(dg)
          subgraphs = nx.weakly_connected_component_subgraphs(dg, copy=False)
          if len(list(subgraphs)) > 1:
            pruned_subgraphs = fido_network.dynamic_pruned_helper(subgraphs, connected_protein_threshold)
            result_lst.extend(pruned_subgraphs)
            break

    return result_lst

  @staticmethod
  def cluster_proteins_from_lists(peptide_spectrum_proteins_prob_charge):
    logger.info('clustering from lists')
    #         given pep --> prots
    #         get
    #         1. prot --> peps
    #         to get
    #         2. peps --> prot
    #         to get prots with identical peps

    prot_to_peps = {}
    for pep,spect,prots,prob,charge in peptide_spectrum_proteins_prob_charge:
      for prot in prots:
        if prot in prot_to_peps: prot_to_peps[prot].update(set([pep]))
        else:                    prot_to_peps[prot] = set([pep])

    peps_to_prots = {}

    for prot,pep_set in list(prot_to_peps.items()):
      pep_set = frozenset(pep_set)
      if pep_set in peps_to_prots: peps_to_prots[pep_set].update(set([prot]))
      else:                        peps_to_prots[pep_set] = set([prot])

    prot_to_protgroup_and_number = {}
    count = 0
    for pep_set,prot_set in list(peps_to_prots.items()):
      if len(prot_set) > 1:
        for prot in prot_set:
          prot_to_protgroup_and_number[prot] = (prot_set,count)
        count += 1

    new_peptide_spectrum_proteins_prob_charge = []
    for pep,spect,prots,prob,charge in peptide_spectrum_proteins_prob_charge:
      new_prots = []
      for prot in prots:
        if prot in prot_to_protgroup_and_number:
          prot_group_set, prot_group_number = prot_to_protgroup_and_number[prot]
          new_prots.append('group_' + str(prot_group_number))
        else: new_prots.append(prot)
      new_peptide_spectrum_proteins_prob_charge.append((pep,spect,new_prots,prob,charge))

    group_set = set()
    for prot in prot_to_protgroup_and_number:
      group_set.add(frozenset(prot_to_protgroup_and_number[prot][0]))

    return new_peptide_spectrum_proteins_prob_charge

  def cluster_proteins(self):
    logger.info('clustering')
    md = multi_dict()
    ### find the proteins that have identical peptide connectivity

    for n in self.node:
      if n['type'] == 'protein':
        md.add(frozenset(self[n]), n)
    ### remove these proteins from the graph and replace them with
    ### a single group that contains all of them
    removal_protein_set = set()
    group_number = 0
    group_set = set()
    for pep_set in md:
      prot_list = list(md[pep_set])
      if len(prot_list) > 1:
        group_name = 'protein_group_'+str(group_number)
        group_set.add(group_name)
        prot_group_node = iid_probability_node(identifier({'id':group_name, 'type':'protein', 'protein_group':frozenset(prot_list)}))
        for pep_id in pep_set:
          pep_node = self.get_node_from_id(pep_id)
          ### note: this will need to be modified if more
          ### complicated edge labes are used
          self.add_edge(prot_group_node, pep_node, label='input')
        removal_protein_set.update(set(prot_list))
        group_number += 1
    self.remove_nodes_from([ self.get_node_from_id(n_id) for n_id in removal_protein_set ])
    return set([ rp['id'] for rp in removal_protein_set ])

  @staticmethod
  def wrap_peptide_id(peptide, assumed_charge, merge_peptide_charge_states):
    peptide_id = {'id':peptide, 'type':'peptide'}
    if not merge_peptide_charge_states:
      peptide_id['charge'] = assumed_charge
    return identifier(peptide_id)

  @staticmethod
  def change_isoleucine_to_leucine(pepStr):
    ### this will ensure that peptides that are identical aside
    ### from I -> L or L -> I changes will be treated as
    ### degenerate
    return pepStr.replace('I','L')

  @staticmethod
  def remove_bounding_amino_acids(pepStr):
    if len(pepStr) > 1 and pepStr[1] == '.':
      return pepStr[2:-2]
    else: return pepStr

  def load_from_pivdo2(self, filename, merge_peptide_charge_states=True):
    bayesian_network.__init__(self)
    ### load the pivdo2 file
    file = open(filename, 'r')
    pivdo2_lines = file.readlines()
    ### get the charge priors and edge information
    charge_priors = fido_network.charge_priors_from_pivdo2(pivdo2_lines)

    peptide_spectrum_proteins_prob_charge = fido_network.get_graph_data_from_pivdo2(pivdo2_lines)

    self.make_graph_from_charge_priors_and_lists( charge_priors = charge_priors, peptide_spectrum_proteins_prob_charge = peptide_spectrum_proteins_prob_charge, merge_peptide_charge_states = merge_peptide_charge_states )

  @staticmethod
  def charge_priors_from_pivdo2(pivdo_lines):
    ### fixme error: this makes pruning use peptideprophet scores
    ### by making the likelihood equal the probability; fix this
    ### before you do any inference!
    return dict.fromkeys(list(range(-1,10)),0.1)

    ### use -1 as unspecified charge
    #charge_priors = { -1:0.1 }
    for line in pivdo_lines:
      line_list = line.split()
      if line_list[0] == 'd':
        charge = int(line_list[1])
        probability = float(line_list[2])
        charge_priors[charge] = probability
    return charge_priors

  @staticmethod
  def get_graph_data_from_pivdo2(pivdo_lines):
    peptide_spectrum_proteins_prob_charge = []

    spectrum_count = 0
    state = ('e')
    for line in pivdo_lines:
      line_list = line.split()
      line_type = line_list[0]
          
      if line_type == 'd': continue

      if line_type not in state:
        raise Exception('Wanted line_type ' + ''.join(state) + ' and got ' + line_type + ' one line: ' + line)

      if line_type == 'e':
        if state == ('r','p','e'):
          peptide_spectrum_proteins_prob_charge.append( (peptide, 'spectrum_'+str(spectrum_count), protein_list, None, charge) )
          spectrum_count += 1

        peptide = line_list[1]
        ### use -1 as unknown or unspecified charge
        charge = -1
        protein_list = []
        state = ('c', 'r')

      elif line_type == 'c':
        charge = int(line_list[1])
        state = ('r')
      elif line_type == 'r':
        protein = line_list[1]
        protein_list.append(protein)
        state = ('r','p','e')
      elif line_type == 'p':
        probability = float(line_list[1])
        state = ('e')
      else:
        raise Exception('Unrecognized line_type in pivdo2 ' + line_type + ' on line ' + line)
      
      if line_type == 'p':
        peptide_spectrum_proteins_prob_charge.append( (peptide, 'spectrum_'+str(spectrum_count), protein_list, probability, charge) )
        spectrum_count += 1

    return peptide_spectrum_proteins_prob_charge

  def make_graph_from_charge_priors_and_lists(self, charge_priors, peptide_spectrum_proteins_prob_charge, merge_peptide_charge_states):
    for peptide, spectrum, protein_list, peptide_probability, assumed_charge in peptide_spectrum_proteins_prob_charge:

      if self.clean_peptide_id:
        peptide = fido_network.change_isoleucine_to_leucine(peptide)
        peptide = fido_network.remove_bounding_amino_acids(peptide)
      
      peptide_id = fido_network.wrap_peptide_id(peptide, assumed_charge, merge_peptide_charge_states)

      ### add protein --> peptide edges
      pep_node = noisy_or_node( peptide_id )
      for prot in protein_list:
        protein_id = identifier({'id':prot, 'type':'protein'})
        prot_node = iid_probability_node(protein_id)
        if prot_node not in self:
          self.add_node( prot_node )
        self.add_edge( prot_node, pep_node, label='input' )

      if peptide_probability != None:
        ### compute the likelihoods from the prior for that particular charge state
        cp = charge_priors[assumed_charge]

        likelihood_given_present = peptide_probability / cp
        likelihood_given_absent = (1-peptide_probability) / (1-cp)
      
        ### add peptide --> spectrum edges
        spectrum_id = identifier({'id':spectrum, 'type':'spectrum'})

        spectrum_node = predecessor_table_node( spectrum_id,  {True:{True: likelihood_given_present, False: likelihood_given_absent}} )
        self.add_edge(pep_node, spectrum_node, label='input')
      else:
        logger.info('Warning: found search_result {} with no spectrum matches'.format(peptide))
        ### (it may match be a non-best match to a spectrum or
        ### may be found elsewhere in the graph with its own
        ### spectrum match)

  def remove_all_but_maximum_likelihood_spectrum(self):
    pep_ids = [ n for n in self if n['type'] == 'peptide' ]
    for pi in pep_ids:
      spectrum_ids = [ sp for sp in self.successors(pi) ]
      if len(spectrum_ids) == 0:
        logger.error('error: peptide', pi['id'], 'has no spectrum matches')
      spectrum_nodes = [ self.get_node_from_id(spect_id) for spect_id in spectrum_ids ] 
      pep_node = self.get_node_from_id(pi)
      likelihoods_present_peptide = [ sp.likelihood_table[True][True] for sp in spectrum_nodes]
      highest_likelihood_spectrum_node = spectrum_nodes[ index_max(likelihoods_present_peptide) ]
      self.remove_nodes_from([ self.get_node_from_id(n_id) for n_id in self.successors(pi) ])
      self.add_edge(pep_node, highest_likelihood_spectrum_node, label='input')

  def multi_file_load(self, fname_list, filetype = 'pivdo', **Kwargs):
    for fname in fname_list:
      logger.info('loading from {}'.format(fname))
      sub_fn = fido_network(**Kwargs)
      ### use the command line value for filetype
      sub_fn.load(fname, filetype, **Kwargs)
      sub_fn.remove_all_but_maximum_likelihood_spectrum()
      self = bayesian_network_union(self, sub_fn)

    logger.info('prots: {}'.format(len([ n for n in self if n['type'] == 'protein' ])))
    logger.info('peps: {}'.format(len([ n for n in self if n['type'] == 'peptide' ])))

  def load(self, fname, filetype = 'pivdo', **Kwargs):
    if filetype == 'pivdo':
      self.load_from_pivdo2(fname)
    else:
      logger.error('Error: unrecognized filetype in fido_network::load:{}'.format(filetype))
      sys.exit(2)

  def load_dataframe(self, df, merge_peptide_charge_states=True, **Kwargs):
    bayesian_network.__init__(self)

    # get charge info -- dummied for now. see charge_priors_from_pivdo2
    #charge_priors = dict.fromkeys(list(range(-1,10)), 1e-6)
    charge_priors = dict.fromkeys(list(range(-1,10)), 0.1)

    # get edge info
    # each element of this list is a tuple describing the PSM
    peptide_spectrum_proteins_prob_charge = []

    sequence = ''
    prot_list = []
    prob = 0.0
    charge = -1 # dummy charge for now
    spectrum_count = 0

    #prot_cols = ['Leading proteins', 'Proteins']
    #prot_cols = ['Leading proteins']
    prot_cols = [self.proteins_column]

    for i in range(0, df.shape[0]):
      sequence = df[self.sequence_column][i]
      prob = 1 - df[self.error_prob_column][i]

      prot_list = []
      for col in prot_cols:
        prots = df[col].loc[i]
        if pd.isnull(prots): continue
        #prot_list = prot_list + [str.join('_', prot.split('|')[0:2]) for prot in prots.split(';')]
        prot_list = prot_list + [prot for prot in prots.split(self.protein_delimiter)]

      prot_list = np.unique(prot_list)
      if len(prot_list) < 1: continue

      peptide_spectrum_proteins_prob_charge.append( (sequence, 'spectrum_'+str(spectrum_count), prot_list, prob, charge) )
      spectrum_count += 1

    self.make_graph_from_charge_priors_and_lists(\
      peptide_spectrum_proteins_prob_charge=peptide_spectrum_proteins_prob_charge,
      charge_priors=charge_priors,
      merge_peptide_charge_states=merge_peptide_charge_states)

  def load_from_dataframes(self, df_list, **Kwargs):
    for df in df_list:
      sub_fn = fido_network(**Kwargs)
      sub_fn.load_dataframe(df, **Kwargs)
      #sub_fn.remove_all_but_maximum_likelihood_spectrum()
      self = bayesian_network_union(self, sub_fn)
    
    logger.info('prots: {}'.format(len([ n for n in self if n['type'] == 'protein' ])))
    logger.info('peps: {}'.format(len([ n for n in self if n['type'] == 'peptide' ])))


def set_gab(gamma, alpha, beta):
  iid_probability_node.gamma = gamma
  noisy_or_node.alpha = alpha
  noisy_or_node.beta = beta

def comp_roc_fdr(posteriors, decoy_tag='REV__'):
  fps = [0]
  tps = [0]
  est_fdr_list = [0]
  emp_fdr_list = [0]

  fp_count = 0
  tp_count = 0
  total_fdr = 0.0
  est_fdr = 0.0
  emp_fdr = 0.0

  num_decoys = 0
  num_targets = 0

  last_prob = -1
  for i in range(0, posteriors.shape[0]):
      prob = posteriors.index[i]
      prots = posteriors.iloc[i]
      #print(prots)
      fp_change = np.sum(decoy_tag in prot for prot in prots)
      tp_change = len(prots) - fp_change
      #print(fp_change, tp_change)
      
      num_decoys += fp_change
      num_targets += tp_change
      
      if tp_change > 0 and fp_change > 0:
          tp_change = 0
          fp_change = 0
      elif tp_change > 0:
          tp_change = 1
      elif fp_change > 0:
          fp_change = 1
      
      if i != 0 and prob != last_prob:
          fps.append(fp_count)
          tps.append(tp_count)
          
          if est_fdr > est_fdr_list[-1]:
              est_fdr_list.append(est_fdr)
              emp_fdr_list.append(emp_fdr)
      
      fp_count += fp_change
      tp_count += tp_change
      
      total_fdr += ((1-prob) * (fp_change + tp_change))
      est_fdr = (total_fdr / (fp_count + tp_count))
      emp_fdr = (float(fp_count) / (fp_count + tp_count))

      last_prob = prob

  fps.append(fp_count)
  tps.append(tp_count)

  fps.append(num_decoys)
  tps.append(num_targets)

  est_fdr_list = np.array(est_fdr_list)
  emp_fdr_list = np.array(emp_fdr_list)

  return (fps, tps, est_fdr_list, emp_fdr_list)

def antiderivative_at(m, b, xVal):
    return (m * xVal * xVal / 2.0) + (b * xVal)

def squared_antiderivative_at(m, b, xVal):
    u = m*m
    v = 2*m*b
    t = b*b
    
    return (u * xVal * xVal / 3.0) + (v * xVal * xVal / 2.0) + (t * xVal)

def area(x1, y1, x2, y2, max_x):
    m = (y2-y1) / (x2-x1)
    b = y1-(m*x1)
    result = antiderivative_at(m, b, np.min([max_x, x2])) - antiderivative_at(m, b, x1)
    if result < 0.0:
        logger.error('area: {}\n{} {} {} {}'.format(result, m, b, x1, x2))
    return result

def squared_area(x1, y1, x2, y2, max_x):
    if x2 < x1: return 0.0
    
    m = ((y2-y1) / (x2-x1))
    b = (y1-(m*x1))
    
    result = (squared_antiderivative_at(m, b, np.min([max_x, x2])) - \
              squared_antiderivative_at(m, b, x1))
    return result

def roc_N(fps, tps, N=50):
  roc_N = 0.0

  if fps[-1] < N:
    logger.warning('Warning: There are not enough false positives; needed {} and was only given {}. Will proceed using largest available value'.format(N, fps[-1]))
    N = fps[-1]
      
  for i in range(0, len(fps)-1):
    if fps[i] >= N: break
    if fps[i] != fps[i+1]:
      current_area = area(fps[i], tps[i], fps[i+1], tps[i+1], N)
      roc_N += current_area
      
  return roc_N / (N * tps[-1])

def fdr_divergence(est_fdr, emp_fdr, thresh=0.1):
  # FDR divergence
  diff = est_fdr - emp_fdr
  tot = 0.0

  for i in range(0, len(diff)-1):
    if est_fdr[i] >= thresh:
      if i == 0:
        tot = np.inf
      break
    
    tot += squared_area(est_fdr[i], diff[i], est_fdr[i+1], diff[i+1], est_fdr[i+1])

  x_range = np.min([thresh, est_fdr[-2]]) - est_fdr[0]

  if np.isinf(tot):
    logger.error('infinite FDR divergence')
    return tot

  return tot / x_range


def run_internal(df, parameter_map):

  if parameter_map['gamma'] is None or parameter_map['alpha'] is None or \
    parameter_map['beta'] is None:

    logger.info('gamma, alpha, beta, not defined. choosing best parameters...')

    logger.info('estimating parameters with accuracy level {}'.format(\
      parameter_map['parameter_accuracy']))
    if parameter_map['parameter_accuracy'] > 1:
      logger.info('sampling smaller dataframe from inputs...')
      sample_size = -1
      if parameter_map['parameter_accuracy'] == 2:
        sample_size = 300
      elif parameter_map['parameter_accuracy'] == 3:
        sample_size = 100

      dfs = df.sample(n=sample_size, random_state=1).reset_index(drop=True)
    else:
      dfs = df

    logger.info('loading into graph network...')
    fn = fido_network(**parameter_map)
    fn.load_from_dataframes(dfs_list, **parameter_map)

    if parameter_map['omit_clean_peptide_name']:
      logger.info('keeping original peptide sequences...')
    else:
      logger.info('cleaning peptide sequences...')

    if parameter_map['all_psms']:
      logger.info('including all PSMs...')
    else:
      logger.info('trimming down to best spectrum match...')
      fn.remove_all_but_maximum_likelihood_spectrum()

    logger.info('pruning...')
    #fn = fido_network.prune_low_scoring_peptides(fn, 1e-2)
    fn = fido_network.dynamic_pruned(fn, **parameter_map)
    fn.init_parameter_names_to_nodes_maps()

    gamma_search = [0.1, 0.5, 0.9];
    alpha_search = [0.01, 0.04, 0.09, 0.16, 0.25, 0.36];
    beta_search = [0.0, 0.01, 0.025, 0.05];

    logger.info('searching parameters for {} iterations'.format(
      len(gamma_search)*len(alpha_search)*len(beta_search)))

    gamma_best = 0.5
    alpha_best = 0.1
    beta_best = 0.01

    best_objective = -10000000
    iter_number = 1

    for i, gamma in enumerate(gamma_search):
      for j, alpha in enumerate(alpha_search):
        for k, beta in enumerate(beta_search):
          # set gamma, alpha, beta
          set_gab(gamma, alpha, beta)

          prots_to_posteriors = fn.FidoMarginalization_Inference()
          # convert to dataframe
          ids = list(prots_to_posteriors.keys())
          probs = [prots_to_posteriors[_id] for _id in ids]
          ids = [_id['id'] for _id in ids]
          posteriors = pd.DataFrame({'prot': ids, 'prob': probs})\
            .sort_values('prob', ascending=False).reset_index(drop=True)
          # group proteins by posterior probability
          posteriors = posteriors.groupby('prob')['prot'].apply(lambda x: x.values)\
            .sort_index(ascending=False)

          fps, tps, est_fdr, emp_fdr = comp_roc_fdr(posteriors, decoy_tag=parameter_map['decoy_tag'])
          roc50 = roc_N(fps, tps, N=50)
          fdr_mse = fdr_divergence(est_fdr, emp_fdr, thresh=0.1)
          _lambda = 0.15

          current_objective = (_lambda * roc50) - ((1 - _lambda) * fdr_mse)
          if current_objective > best_objective:
            best_objective = current_objective
            gamma_best = gamma
            alpha_best = alpha
            beta_best = beta

          logger.info('iteration {} || gamma: {}, alpha: {}, beta: {}'.format(
            iter_number, gamma, alpha, beta))
          logger.info('                roc50: {:.4f}, fdr_mse: {:.4f}, objective: {:.4f}'.format(\
            roc50, fdr_mse, current_objective))

          iter_number += 1

    logger.info('best objective: {:.4f}, with parameters {}, {}, {}'.format(best_objective, gamma_best, alpha_best, beta_best))

    # set gamma, alpha, beta
    parameter_map['gamma'] = gamma_best
    parameter_map['alpha'] = alpha_best
    parameter_map['beta'] = beta_best

  logger.info('running with parameters gamma: {}, alpha: {}, beta: {}'.format(
    parameter_map['gamma'], parameter_map['alpha'], parameter_map['beta']))

  # run with the best parameters
  set_gab(parameter_map['gamma'], parameter_map['alpha'], parameter_map['beta'])

  fn = fido_network(**parameter_map)
  
  logger.info('loading into graph network...')
  fn.load_from_dataframes([df], **parameter_map)

  if parameter_map['all_psms']:
    logger.info('including all PSMs...')
  else:
    logger.info('trimming down to best spectrum match...')
    fn.remove_all_but_maximum_likelihood_spectrum()

  if parameter_map['group_proteins']:
    logger.info('clustering proteins...')
    fn.cluster_proteins()

  logger.info('pruning...')
  if parameter_map['prune_low_scores']:
    logger.info('first pruning low scoring peptides (confidence < 1e-2)')
    fn = fido_network.prune_low_scoring_peptides(fn, 1e-2)

  fn = fido_network.dynamic_pruned(fn, **parameter_map)
  fn.init_parameter_names_to_nodes_maps()
  
  logger.info('running fido inference...')
  prots_to_posteriors = fn.FidoMarginalization_Inference()
  #pprint(prots_to_posteriors)

  ids = list(prots_to_posteriors.keys())
  probs = [prots_to_posteriors[_id] for _id in ids]
  ids = [_id['id'] for _id in ids]
  
  out_df = pd.DataFrame({'Protein': ids, 'Probability': probs})\
    .sort_values('Probability', ascending=False).reset_index(drop=True)

  #print(np.percentile(out_df['Probability'], np.arange(10, 100, 10)))

  out_df['Error Probability'] = 1 - out_df['Probability']
  out_df['q-value'] = np.cumsum(out_df['Error Probability'].values) / np.arange(1, out_df.shape[0]+1, 1)

  # save output from fido to a file, protein_fdr.txt
  logger.info('Writing Fido protein inference output to file: {}'.format(os.path.join(parameter_map['output'], 'protein_fdr.txt')))
  out_df.to_csv(os.path.join(parameter_map['output'], 'protein_fdr.txt'), 
    sep='\t', index=False)

  # attach PSM razor protein FDR to the input dataframe
  logger.info('Attaching razor protein FDR to output')
  fdr_series = pd.Series(out_df['q-value'].values, index=out_df['Protein'].values)
  df['razor_protein_fdr'] = df[parameter_map['leading_protein_column']].map(fdr_series)

  #print(np.sum(df['razor_protein_fdr'] < 0.1), df.shape[0])

  return df
  


def pd_main():

  # initialize logger
  init_logger(True, 'fido.log', log_to_file=False)

  parser = argparse.ArgumentParser()  
  parser.add_argument('-l', '--connected-protein-threshold', type=int, default=14,
    help='log2 of maximum number of subgraph connected states')
  parser.add_argument('-p', '--omit-clean-peptide-name', action='store_true', default=False,
    help='omit cleaning the peptide names')
  parser.add_argument('-a', '--all-psms', action='store_true', default=False,
    help='use all PSM matches instead the best one')
  parser.add_argument('-g', '--group-proteins', action='store_true', default=False,
    help='use protein group level inference')
  parser.add_argument('-s', '--prune-low-scores', action='store_true', default=False,
    help='prune low scoring PSMs (threshold set at 0.01)')
  parser.add_argument('-c', '--parameter-accuracy', type=int, default=3,
    help='set start parameter\'s accurary level (1-3)\n1 = best (slower)\n2 = relaxed (faster)\n3 = sloppy (very fast)')

  parser.add_argument('--gamma', type=float, default=None,
    help='protein prior probability')
  parser.add_argument('--alpha', type=float, default=None,
    help='peptide emission probability')
  parser.add_argument('--beta', type=float, default=None,
    help='spurious peptide identification probability')

  parser.add_argument('--proteins-column', type=str, default='Proteins',
    help='column name for the proteins of each PSM')
  parser.add_argument('--protein-delimiter', type=str, default=';',
    help='delimiter character separating proteins in the proteins column')
  parser.add_argument('--leading-protein-column', type=str, default='Leading razor protein',
    help='column name for the leading (razor) protein of each PSM')
  parser.add_argument('--decoy-tag', type=str, default='REV__',
    help='expression used to deliminate decoy proteins from target proteins')
  parser.add_argument('--sequence-column', type=str, default='Sequence',
    help='column name for the peptide sequence of each PSM')
  parser.add_argument('--error-prob-column', type=str, default='PEP',
    help='column name for the error probability of each PSM')

  parser.add_argument('-o', '--output', type=str, default=os.getcwd(), help='Folder to output protein_fdr list and intermediate fido files')

  args = parser.parse_args()

  #parameter_map = {
  #  'connected_protein_threshold': 14,
  #  'gamma': 0.5, 'alpha': 0.1, 'beta': 0.01 }
  parameter_map = vars(args)

  logger.info('reading...')
  df = pd.read_csv('/gd/MS/SCoPE/SQC/SQC77/evidence.txt', sep='\t', low_memory=False)
  #df = pd.read_csv('~/git/RTLib/Alignments/SQC2_20180812_1/ev_updated.txt', 
  #  sep='\t', low_memory=False)

  run_internal(df, parameter_map)

def real_main(argv):

  # initialize logger
  init_logger(True, 'fido.log', log_to_file=False)

  try:
    opts, args = getopt.getopt(argv, 'f:gc:m:', ['filetype=', 'group_proteins', 'connected_protein_threshold=', 'marginalization_mode='])
  except getopt.GetoptError:
    print('cmd line error')
    sys.exit(2)

  ### this converts any short option (like -f) into the appropriate
  ### long option (--filetype)
  shorthand_to_longhand = {'-f':'--filetype', '-g':'--group_proteins', '-c':'--connected_protein_threshold', '-m':'--marginalization_mode'}

  longhand_opts = []
  for flag, value in opts:
    if flag in shorthand_to_longhand:
      longhand_opts.append( (shorthand_to_longhand[flag], value) )
    else:
      longhand_opts.append( (flag, value) )

  parameter_map = { \
    'marginalization_mode': 'fido', 
    'filetype': 'pivdo', 
    'connected_protein_threshold': 14 }

  #parameter_map = { }
  for flag, value in longhand_opts:
    ### strip off the leading '--'
    parameter_map[flag[2:]] = value

  print('parameter_map')
  print(parameter_map)

  fn = fido_network(**parameter_map)
  print('args',args)
  print('reading...')

  fn = fido_network(**parameter_map)
  fn.multi_file_load(args, **parameter_map)

  print('trimming down to best spectrum match...')
  fn.remove_all_but_maximum_likelihood_spectrum()

  if 'group_proteins' in parameter_map:
    print('clustering proteins')
    fn.cluster_proteins()
  
  print('pruning...')

  #fn = fido_network.prune_low_scoring_peptides(fn, 1e-2)
  fn = fido_network.dynamic_pruned(fn, **parameter_map)

  fn.init_parameter_names_to_nodes_maps()

  if parameter_map['marginalization_mode'] == 'fido':
    print('fido inference')
    prots_to_posteriors = fn.FidoMarginalization_Inference()
  else:
    print('Error: unrecognized marginalization_mode = ', parameter_map['marginalization_mode'])

  pprint(prots_to_posteriors)

main = real_main

if __name__ == '__main__':
  main(sys.argv[1:])





#!/usr/bin/env python3
# coding: utf-8

# converter.py - converters from search-engine inputs to compatible outputs
# for alignment and update

import argparse
import logging
import numpy as np
import os
import pandas as pd
import re

from functools import reduce
from dart_id.exceptions import ConfigFileError, FilteringError
from dart_id.helper import add_global_args, read_config_file, init_logger

logger = logging.getLogger('root')

# all filter funcs take in the df, config object, and the filter object
# as inputs, and output the filter vector (True/False), where True means
# to filter out that particular row

def filter_exclude_filename(df, config, _filter):
  # remove experiments from blacklist
  if 'expr' not in _filter or len(_filter['expr']) == 0:
    raise ConfigFileError('Experiment exclusion by raw file name provided, but exclusion expression is defined incorrectly. Please provide a regular expression, with the \"expr\" key, to match against raw file names, that determines whether or not that raw file is included in the alignment procedure.')
  
  # see if any raw file names match the user-provided expression
  exclude_exps = list(filter(lambda x: re.search(r'' + _filter['expr'] + '', x), 
    df['raw_file'].unique()))
  # remove excluded rows from the sparse dataframe,
  # but keep them in the original data frame, so that we can stitch together
  # the final output later
  exclude_exps = df['raw_file'].isin(exclude_exps).values
  df = df[~exclude_exps]
  logger.info('Filtering out {} observations matching \"{}\"'.format(np.sum(exclude_exps), _filter['expr']))

  return exclude_exps

def filter_include_filename(df, config, _filter):
  # exclusively include experiments from whitelist
  if 'expr' not in _filter or len(_filter['expr']) == 0:
    raise ConfigFileError('Experiment inclusion by raw file name provided, but inclusion expression is missing or defined incorrectly. Please provide a regular expression, with the \"expr\" key, to match against raw file names, that determines whether or not that raw file is included in the alignment procedure.')
  
  # get matches for this expression
  include_exps = list(filter(lambda x: re.search(r'' + _filter['expr'] + '', x), df['raw_file'].unique()))
  # only keep rows that are in these raw file matches
  include_exps = df['raw_file'].isin(include_exps).values
  df = df[include_exps]
  logger.info('Keeping {} observations out of {} matching inclusion expression \"{}\"'.format(np.sum(include_exps), 
      df.shape[0], _filter['expr']))

  # filter out the opposite of the included experiments
  return ~include_exps

def filter_uniprot_exclusion_list(df, config, _filter):
  """
  Filter proteins from exclusion list using UniProt IDs
  """

  exclusion_list = []

  # parse exclusion list
  # if exclusion_list param is a path, then load the IDs from that path
  if 'file' in _filter and _filter['file'] is not None:
    # load UniProt IDs from file line-by-line
    # first expand user or any vars
    _filter['file'] = os.path.expanduser(_filter['file'])
    _filter['file'] = os.path.expandvars(_filter['file'])

    # open the exclusion list file and read in the UniProt IDs, line by line
    try:
      with open(_filter['file'], 'r') as f:
        logger.info('Loading UniProt IDs from exclusion list file {} ...'.format(_filter['file']))
        exclusion_list = [line.rstrip('\n') for line in f]
        logger.info('Loaded {} proteins from exclusion list.'.format(len(exclusion_list)))
    except EnvironmentError:
      raise ConfigFileError('Exclusion list file {} not found. Please provide a path to a file with UniProt IDs separated by line'.format(_filter['file']))

  elif 'list' in _filter and len(_filter['list']) > 0:
    # load UniProt IDs from the configuration file
    exclusion_list = _filter['list']
    logger.info('Loading {} UniProt IDs from exclusion list as defined in config file'.format(len(exclusion_list)))
  else:
    raise ConfigFileError('No exclusion list file or list of UniProt IDs provided. Please provide a path to a file with UniProt IDs separated by line with the \"file\" key, or provide a python list of UniProt IDs with the \"list\" key. If not using a UniProt ID exclusion list, then comment out the \"uniprot_exclusion\" key from the filter list.')

  # filter exclusion list
  if len(exclusion_list) > 0:
    logger.info('UniProt IDs from exclusion list: {}'.format(exclusion_list))

    # we could only match the excluded IDs to the razor protein,
    # but we can be more strict and match the blacklisted IDs to the entire protein
    # string, containing all possible proteins
    pat = reduce((lambda x, y: x + '|' + y), exclusion_list)
    blacklist_filter = df['proteins'].str.contains(pat)
    blacklist_filter[pd.isnull(blacklist_filter)] = False

    logger.info('Filtering out {} PSMs from the exclusion list'.format(np.sum(blacklist_filter)))
    return blacklist_filter
  else:
    raise ConfigFileError('Exclusion list found and loaded, but no UniProt IDs found. Check the format of the file, or the list in the config file.')

def filter_contaminant(df, config, _filter):
  """
  Filter contaminants, as marked by the search engine
  Looking for a contaminant tag in the leading_protein column
  """

  # load the tag in from the config file
  CON_TAG = _filter['tag']

  if CON_TAG is None:
    raise ConfigFileError('Contaminant filter has no tag specified. Please provide an expression with the \"tag\" key that matches a contaminant tag from the search engine output. If you don\'t want to apply this filter, then comment out the \"contaminant\" key from the filter list.')
  elif type(CON_TAG) is not str:
    raise ConfigFileError('Contaminant tag {} is not of type string. Please provide an expression with the \"tag\" key that matches a contaminant tag from the search engine output.'.format(CON_TAG))
  elif len(CON_TAG) <= 0:
    raise ConfigFileError('Contaminant tag empty. Please provide an expression with the \"tag\" key that matches a contaminant tag from the search engine output.')

  # search for the tag in the 'proteins' column
  filter_con = df['proteins'].str.contains(CON_TAG)
  filter_con[pd.isnull(filter_con)] = False

  logger.info('Filtering out {} PSMs as contaminants with tag \"{}\"'.format(np.sum(filter_con), CON_TAG))
  return filter_con

def filter_decoy(df, config, _filter):
  """
  Filter decoys, as marked by the search engine
  Looking for a decoy tag in the leading_protein column
  """

  # load the tag in from the config file
  REV_TAG = _filter['tag']

  if REV_TAG is None:
    raise ConfigFileError('Decoy filter has no tag specified. Please provide an expression with the \"tag\" key that matches a decoy tag from the search engine output. If you don\'t want to apply this filter, then comment out the \"decoy\" key from the filter list.')
  elif type(REV_TAG) is not str:
    raise ConfigFileError('Decoy tag {} is not of type string. Please provide an expression with the \"tag\" key that matches a decoy tag from the search engine output.'.format(REV_TAG))
  elif len(REV_TAG) <= 0:
    raise ConfigFileError('Decoy tag empty. Please provide an expression with the \"tag\" key that matches a decoy tag from the search engine output.')

  # search for the tag in the 'leading protein' column
  filter_rev = df['leading_protein'].str.contains(REV_TAG)
  filter_rev[pd.isnull(filter_rev)] = False

  logger.info('Filtering out {} PSMs as decoys with tag \"{}\"'.format(np.sum(filter_rev), REV_TAG))
  return filter_rev

def filter_retention_length(df, config, _filter):
  """
  Filter by retention length, which is a measure of the peak width
  during chromatography.
  """

  # input checks
  if 'value' not in _filter or _filter['value'] is None:
    raise ConfigFileError('No value provided to the retention_length filter. If you don\'t want to apply this filter, then comment out the \"retention_length\" key from the filter list. Or, please provide a threshold with the \"value\" key, and specify whether or not this is an absolute time in minutes with \"dynamic\"=false or \"dynamic\"=true if the filter is a fraction of the experiment run-time. If \"dynamic\"=true, then \"value\" must be between 0 and 1.')
  if 'dynamic' not in _filter or type(_filter['dynamic']) is not bool:
    raise ConfigFileError('Incorrect value provided to the \"dynamic\" field of the retention_length filter. Please provide a bool, either true or false. If \"dynamic\"=false, then \"value\" is the threshold in minutes. If \"dynamic\"=true, then \"value\" must be between 0 and 1.')
  
  if _filter['dynamic']:
    # use the dynamic filter, where the value is a proportion
    # of the max RT (the run-time) of that raw file
    
    # only allow values between 0 and 1 for the dynamic filter
    if _filter['value'] > 1 or _filter['value'] <= 0:
      raise ConfigFileError('Dynamic retention_length filter {} is above 1 or below 0. Please provide a number between 0 and 1, which is the fraction of the max RT for each experiment. e.g., 0.01 means that 1%% of the max RT will be used as the retention_length threshold.'.format(_filter['value']))

    logger.info('Using dynamic retention length of {} * run-time (max RT) for each experiment'.format(_filter['value']))

    # get the max RT for each raw file, reindex to the same dimension as the
    # retention_length column, and then multiply by the filter value
    max_rts = df.groupby('raw_file')['retention_time'].max().values
    
    filter_rtl = max_rts[df['raw_file'].map({ 
      ind: val for val, ind in enumerate(np.sort(df['raw_file'].unique()))
    })] * _filter['value']

    filter_rtl = (df['retention_length'] > filter_rtl)

  else:
    # use a constant filter for the retention length
    logger.info('Using constant retention length (in RT) of {} for all raw files.'.format(_filter['value']))

    # only allow values between 0 and max(RT)
    if _filter['value'] <= 0 or _filter['value'] > np.max(df['retention_time']):
      raise ConfigFileError('retention_length filter {} is not defined or incorrectly defined. Please provide a decimal number between 0.0 and max(RT).'.format(_filter['value']))
    
    filter_rtl = (df['retention_length'] > _filter['value'])

  if _filter['dynamic']:
    logger.info('Filtering out {} PSMs with retention length greater than {:.4f} * max(exp_RT) of each raw file.'.format(np.sum(filter_rtl), _filter['value']))
  else:
    logger.info('Filtering out {} PSMs with retention length greater than {:.4f}'.format(np.sum(filter_rtl), _filter['value']))

  return filter_rtl

def filter_smears(df, config, _filter):
  """
  Filter out "smears". even confidently identified PSMs can have bad chromatography,
  and in that case it is unproductive to include them into the alignment.
  In theory, this should be made redundant by the retention length filter, but
  some PSMs still slip through the cracks of that, possibly because the search engine
  cannot adequately track the elution peak?
  """

  # input checking
  if 'value' not in _filter or _filter['value'] is None:
    raise ConfigFileError('No value provided to the \"smears\" filter. If you don\'t want to apply this filter, then comment out the \"smears\" key from the filter list. Or, please provide a threshold with the \"value\" key, and specify whether or not this is an absolute time in minutes with \"dynamic\"=false or \"dynamic\"=true if the filter is a fraction of the experiment run-time. If \"dynamic\"=true, then \"value\" must be between 0 and 1.')
  if 'dynamic' not in _filter or type(_filter['dynamic']) is not bool:
    raise ConfigFileError('Incorrect value provided to the \"dynamic\" field of the \"smears\" filter. Please provide a bool, either true or false. If \"dynamic\"=false, then \"value\" is the threshold in minutes. If \"dynamic\"=true, then \"value\" must be between 0 and 1.')

  logger.info('Determining RT spread of peptides within each experiment...')
  # for each experiment-peptide pair, get the range of retention times
  # this is the step that could take a long time
  # TODO: optimize this?
  smears = df.groupby(['raw_file', 'sequence'])['retention_time'].apply(np.ptp)
  
  if _filter['dynamic']:
    # use the dynamic filter, where the value is a proportion
    # of the max RT (the run-time) of that raw file
    
    if _filter['value'] > 1 or _filter['value'] <= 0:
      raise ConfigFileError('Dynamic smear filter {} is above 1 or below 0. Please provide a number between 0 and 1, which is the fraction of the max RT for each experiment. e.g., 0.01 means that 1%% of the max RT will be used as the retention_length threshold.'.format(_filter['value']))

    logger.info('Using dynamic smear length (in RT) of {:.4f} * run-time (max RT) for each experiment'.format(_filter['value']))

    max_rts = df.groupby('raw_file')['retention_time'].max().values

    smear_pair_inds = smears.index.to_frame()['raw_file'].values
    smear_pair_inds = pd.Series(smear_pair_inds).map({
      ind: val for val, ind in enumerate(np.sort(df['raw_file'].unique())) })

    # get the (raw_file, sequence) tuples for PSMs with a range above the threshold
    smears = smears[smears > (max_rts[smear_pair_inds] * _filter['value'])].index.values

  else:
    # use a constant filter for the retention length
    logger.info('Using constant smear length (in RT) of {:.4f} for all raw files.'.format(_filter['value']))

    if _filter['value'] <= 0:
      raise ConfigFileError('Smear filter {:.4f} is not defined or incorrectly defined. Please provide a decimal number between 0.0 and max(RT).'.format(_filter['value']))

    # get the (exp_id, peptide_id) tuples for PSMs with a range above the threshold
    smears = smears[smears > _filter['value']].index.values
  
  # map the tuples back to the original data frame, and set smears to be excluded
  smears = pd.Series(list(zip(df['raw_file'], df['sequence']))).isin(smears)

  if _filter['dynamic']:
    logger.info('Filtering out {} PSMs with an intra-experiment RT spread greater than {:.4f} * max(exp_RT) for each raw file.'.format(smears.sum(), _filter['value']))
  else:
    logger.info('Filtering out {} PSMs with an intra-experiment RT spread greater than {:.4f}'.format(smears.sum(), _filter['value']))

  return smears.values

# dictionary of all filter functions
filter_funcs = {
  'exclude_filename':   filter_exclude_filename,
  'include_filename':   filter_include_filename,
  'uniprot_exclusion':  filter_uniprot_exclusion_list,
  'contaminant':        filter_contaminant,
  'decoy':              filter_decoy,
  'retention_length':   filter_retention_length,
  'smears':             filter_smears
}

# columns required for each filter to run
# will skip the filter if this column does not exist in the input file
required_cols = {
  'exclude_filename':   [],
  'include_filename':   [],
  'uniprot_exclusion':  ['proteins'],
  'contaminant':        ['proteins'],
  'decoy':              ['leading_protein'],
  'retention_length':   ['retention_length'],
  'smears':             []
}

def convert(df, config):

  cols = []
  col_names = []
  
  # loop thru all columns listed in the config file
  for col in list(config['col_names'].keys()):
    if config['col_names'][col] is None: 
      logger.debug('Column \"{}\" is left empty in the config file. Skipping...'.format(col))
      continue

    # check if the column specified in the config file exists in the df or not
    if config['col_names'][col] not in df.columns:
      # this is probably grounds to kill the program
      raise ConfigFileError('Column {} of value {} not found in the input file. Please check that this column exists, or leave the field for {} empty in the config file.'.format(col, config['col_names'][col], col))

    # keep track of the column and the column name
    cols.append(config['col_names'][col])
    col_names.append(col)

  # take the subset of the input file, and also rename the columns
  dfa = df[cols]
  dfa.columns = col_names

  return dfa


def filter_psms(df, config):
  logger.info('Filtering PSMs...')

  # load the filtering functions specified by the input config
  # types of filter functions depends on what stage of filtering this is:
  # removing observations or merely excluding from alignment
  filters = config['filters']

  # each filter has a specified required column from the dataframe
  # make sure these columns exist before proceeding
  for i, f in enumerate(filters):
    # make sure filter has required field of 'name'
    if 'name' not in f or f['name'] is None or len(f['name']) == 0:
      raise ConfigFileError('Filter {} does not have the required field name or it is empty. Please provide a name (type) for this filter. Available filters are listed in the example config file, under example/config.yaml'.format(str(f)))

    # for each required column in the filter, check if it exists
    for j in required_cols[f['name']]:
      if j not in df.columns:
        raise ConfigFileError('Filter {} required a data column {}, but this was not found in the input dataframe.'.format(f['name'], j))

  # by default, filter out nothing. we'll use binary ORs (|) to
  # gradually add more and more observations to this filter out blacklist
  df['remove'] = np.repeat(False, df.shape[0])

  # run all the filters specified by the list in the input config file
  # all filter functions are passed df, and the run configuration
  # after each filter, append it onto the exclusion master list with a bitwise OR
  # if the filter function returns None, then just ignore it.
  for i, f in enumerate(filters):
    e = filter_funcs[f['name']](df, config, f)
    if e is not None:
      df['remove'] = (df['remove'] | e)

  return df

def process_files(config):

  # create our output data frames
  df_original = pd.DataFrame()
  df = pd.DataFrame()

  # iterate through each input file provided.
  for i, f in enumerate(config['input']):
    # first expand user or any vars
    f = os.path.expanduser(f)
    f = os.path.expandvars(f)

    logger.info('Reading in input file #{} | {} ...'.format(i+1, f))

    # load the input file with pandas
    # 
    # have a variable low memory option depending on the input type.
    # MaxQuant, for example, has a structure that forces pandas out of its
    # optimal low memory mode, and we have to specify it here.
    dfa = pd.read_csv(f, sep='\t', low_memory=config['low_memory'])

    # keep track of where observations came from. this is _not_ the raw file ID
    # but instead the ID from which input file it originated from, so that if
    # we need to split these observations up by input file in the future we can do so
    dfa['input_id'] = i

    # append a copy of dfa into df_original, because the conversion process will heavily
    # modify dfa. we need to keep a copy of the original dataframe in order to append
    # the new columns back onto it later.
    # re-index columns with '[dfa.columns.tolist()]' to preserve the general column order
    df_original = df_original.append(dfa, sort=True)[dfa.columns.tolist()]

    # if this input data already has DART-ID columns in it, then drop them,
    # since they cause problems later
    dart_cols = ['rt_minus', 'rt_plus', 'mu', 'muij', 'sigmaij', 'pep_new', 'exp_id', 'peptide_id', 'stan_peptide_id', 'exclude', 'residual', 'pep_updated', 'q-value']
    # print a warning if we see any
    if np.any(df_original.columns.isin(dart_cols)):
      logger.warning('Columns {} are recognized as DART-ID output columns. Removing these columns before proceeding. In the future, please input original input data files, not output files from DART-ID.'.format(np.array_str(df_original.columns[df_original.columns.isin(dart_cols)])))

      # drop existing dart cols
      for col in dart_cols:
        if col in df_original.columns:
          logger.debug('Removing column {}'.format(col))
          df_original = df_original.drop(col, axis=1)

    logger.info('Converting {} ({} PSMs)...'.format(f, dfa.shape[0]))

    # convert - takes subset of columns and renames them
    dfa = convert(dfa, config)

    # need to reset the input_id after the conversion process
    dfa['input_id'] = i
    # append to master dataframe
    df = df.append(dfa)

  # modify columns?
  # append the ion charge to the sequence
  # also make sure the charge column is specified and exists
  if config['add_charge_to_sequence'] and 'charge' in df.columns:
    logger.info('Appending charge to peptide sequence, to align different charge states separately.')
    df['sequence'] = df['sequence'] + '_' + df['charge'].apply(str)

  # create a unique ID for each PSM to help with stiching the final result together
  # after all of our operations
  df['id'] = range(0, df.shape[0])
  df_original['id'] = range(0, df.shape[0])

  # by default, exclude nothing from the original experiment
  df_original['remove'] = np.repeat(False, df_original.shape[0])

  # if the input already has an 'remove' column, then skip this step
  if 'remove' in config['col_names'] and config['col_names']['remove'] is not None:
    df['remove'] = df['remove'].astype(bool)
  else: # otherwise, run the filters
    df = filter_psms(df, config)

  # apply non-optional filters, PEP threshold and requirement that
  # sequence is observed in at least n experiments (num_experiments)
  
  # first check if PEP threshold is valid
  if 'pep_threshold' not in config or config['pep_threshold'] is None:
    raise ConfigFileError('PEP threshold not defined. Please provide a PEP threshold between 0 and 1 with the \"pep_threshold\" key. If including all PEPs, then set this value to 1.')
  # only allow values between 0 and 1
  if config['pep_threshold'] <= 0 or config['pep_threshold'] > 1:
    raise ConfigFileError('PEP threshold {} is incorrectly defined. Please provide a decimal number between 0.0 and 1.0.'.format(config['pep_threshold']))

  # remove any observations with null pep
  null_pep = pd.isnull(df['pep'])
  if np.sum(null_pep) > 0:
    df['remove'] = ((df['remove']) | (null_pep))
    logger.info('Removing {} PSMs with no PEP entry.'.format(np.sum(null_pep)))

  # check if num_experiments option is valid
  if 'num_experiments' not in config or config['num_experiments'] is None:
    raise ConfigFileError('No value provided to the filter for the number of raw files that a peptide must be observed in. Please provide an integer greater than or equal to 1 with the \"num_experiments\" key.')
  if type(config['num_experiments']) is not int or config['num_experiments'] < 1:
    raise ConfigFileError('Incorrect value of type {} provided to the filter for the number of raw files that a peptide must be observed in. Please provide an integer greater than or equal to 1 with the \"num_experiments\" key.'.format(str(type(config['num_experiments']))))
  num_exps = len(df['raw_file'].unique())
  if config['num_experiments'] > num_exps:
    raise ConfigFileError('Number of experiments filter threshold {} is greater than the number of experiments in the input list. Please provide an integer greater than or equal to 1 and less than the number of experiments with the \"num_experiments\" key.'.format(config['num_experiments']))
  if config['num_experiments'] == 1:
    logger.warning('Filter for number of raw files that a peptide must be observed in is set to 1. The alignment will proceed but this may result in non-informative canonical RTs and high residuals. It is recommended that this parameter is at least 3.')
  
  # count the number of experiments a peptide is observed in, but filter out
  # 1) PSMs removed from previous filters
  # 2) PSMs with PEP > pep_threshold
  exps_per_pep = df[-((df['remove']) | (df['pep'] >= config['pep_threshold']))].groupby('sequence')['raw_file'].unique().apply((lambda x: len(x)))
  # map values to DataFrame. peptides without any value will get NaN,
  # which will then be assigned to 0.
  exps_per_pep = df['sequence'].map(exps_per_pep)
  exps_per_pep[pd.isnull(exps_per_pep)] = 0

  # flag these sequences for removal as well
  logger.info('Removing {} PSMs from peptide sequences not observed confidently in more than {} experiments'.format(np.sum(exps_per_pep < config['num_experiments']), config['num_experiments']))
  df['remove'] = (df['remove'] | (exps_per_pep < config['num_experiments']))

  # check that every experiment has at least n PSMs available for alignment.
  # if not, then exclude them from alignment
  psms_per_exp = df.groupby('raw_file')['remove'].apply(lambda x: np.sum(x < config['pep_threshold']))
  exclude_exps = psms_per_exp.index.values[psms_per_exp < config['min_psms_per_experiment']]
  
  logger.warning('Experiments {} have < {} confident PSMs (PEP < {}) remaining after filtering. All PSMs belonging to these experiments will be excluded from the retention time alignment'.format(np.array_str(exclude_exps), config['min_psms_per_experiment'], config['pep_threshold']))

  # exclude experiments without enough PSMs
  df['remove'] = (df['remove'] | df['raw_file'].isin(exclude_exps))

  # recalculate exps_per_pep, since we removed some experiments and this
  # number will change based on the set of experiments we consider
  logger.info('Recalculating number of confident peptides across experiments...')

  exps_per_pep = df[-((df['remove']) | (df['pep'] >= config['pep_threshold']))].groupby('sequence')['raw_file'].unique().apply((lambda x: len(x)))
  exps_per_pep = df['sequence'].map(exps_per_pep)
  exps_per_pep[pd.isnull(exps_per_pep)] = 0
  logger.info('Additional {} PSMs from peptide sequences not observed confidently in more than {} experiments flagged for removal.'.format(np.sum(exps_per_pep < config['num_experiments']) - np.sum(df['remove']), config['num_experiments']))
  df['remove'] = (df['remove'] | (exps_per_pep < config['num_experiments']))

  ## --------------
  ## DONE FILTERING
  ## --------------

  # flag the observations in df_original that were removed
  df_original['remove'] = df['remove']
  # remove the flagged observations from the dataframe, and reset index
  df = df[df['remove']==False].reset_index(drop=True)

  # map peptide and experiment IDs
  # sort experiment IDs alphabetically - or else the order is by 
  # first occurrence of an observation of that raw file
  
  # if experiment or peptide IDs are already provided, then skip this step
  if 'exp_id' not in config['col_names'] or config['col_names']['exp_id'] is None:
    df['exp_id'] = df['raw_file'].map({
      ind: val for val, ind in enumerate(np.sort(df['raw_file'].unique()))})
  logger.info('{} experiments (raw files) loaded'.format(np.max(df['exp_id'])+1))

  if 'peptide_id' not in config['col_names'] or config['col_names']['peptide_id'] is None:
    df['peptide_id'] = df['sequence'].map({
      ind: val for val, ind in enumerate(df['sequence'].unique())})
  logger.info('{} peptide sequences loaded'.format(np.max(df['peptide_id'])+1))

  # EXCLUSION = PSM does not participate in alignment, but will participate in 
  #             confidence update since the PSM's associated peptide will get
  #             parameters from the alignment. 
  #             This is NOT the same as "remove", which means that the PSM's
  #             associated peptide does not have enough PSMs to participate
  #             in alignment and therefore receive parameters.
  
  # flag non-confident PSMs for exclusion from alignment process
  df['exclude'] = (df['pep'] >= config['pep_threshold'])
  logger.info('Excluding {} / {} ({:.2%}) PSMs from alignment process after filtering at PEP threshold of {}'.format(np.sum(df['pep'] >= config['pep_threshold']), df.shape[0], np.sum(df['pep'] >= config['pep_threshold']) / df.shape[0], config['pep_threshold']))

  # only take the four required columns (+ the IDs) with us
  # the rest were only needed for filtering and can be removed
  df = df[['sequence', 'raw_file', 'retention_time', 'pep', 
           'exp_id', 'peptide_id', 'input_id', 'id', 'exclude']]

  # sort by peptide_id, exp_id
  df = df.sort_values(['peptide_id', 'exp_id'])

  return df, df_original

def main():
  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser)
  args = parser.parse_args()

  # load config file
  # this function also creates the output folder
  config = read_config_file(args)

  # initialize logger
  init_logger(config['verbose'], os.path.join(config['output'], 'converter.log'), config['log_file'])

  # process all input files (converts and filters)
  df, df_original = process_files(config)
  
  #logger.info('{} / {} ({:.2%}) observations pass filters and will be used for alignment'.format(np.sum(~df['exclude']), 
  #    df_original.shape[0], np.sum(~df['exclude']) / df_original.shape[0]))

  # if no output is specified, throw an error
  if config['save_combined_output'] == False and config['save_separate_output'] == False:
    raise ConfigFileError('No output format specified. Either set \"save_combined_output\" to true, or set \"save_separate_output\" to true.')

  # write to file
  if config['save_combined_output']:
    # if combining input files, then write to one combined file
    out_path = os.path.join(config['output'], config['combined_output_name'])
    logger.info('Combining input file(s) and writing adjusted data file to {} ...'.format(out_path))
    df.to_csv(out_path, sep='\t', index=False)
  
  if config['save_separate_output']:
    # if keeping input files separate, then use 'input_id' to retain the
    # order in which the input files were passed in
    logger.info('Saving output to separate files...')
    for i, f in enumerate(config['input']):
      out_path = os.path.join(config['output'], 
        os.path.splitext(os.path.basename(f))[0] + config['output_suffix'] + \
          '_' + str(i) + '.txt')
      logger.info('Saving input file {} to {}'.format(i, out_path))
      df_a = df.loc[df['input_id'] == i]
      df_a.to_csv(out_path, sep='\t', index=False)

if __name__ == '__main__':
  main()



# coding: utf-8

# models.py - define models and their parameters

import logging
import numpy as np
import os
import pandas as pd

from dart_id.exceptions import ConfigFileError
from scipy.stats import norm, lognorm, laplace

logger = logging.getLogger('root')

def generate_inits_linear(dff, config):

  exp_names = np.sort(dff['raw_file'].unique())
  num_experiments = len(exp_names)
  num_peptides = dff['stan_peptide_id'].max() + 1

  # get the average retention time for a peptide, weighting by PEP
  def get_mu(x):
    weights = ((1 - x['pep'].values) - (1 - config['pep_threshold'])) / config['pep_threshold']
    return np.sum(x['retention_time'].values * weights) / np.sum(weights)

  # apply the get_mu function on all peptides and add some distortion
  mu_init = dff.groupby('stan_peptide_id')[['pep', 'retention_time']].apply(get_mu).values + np.random.normal(0, config['rt_distortion'], num_peptides)

  # negative or very low retention times not allowed. floor at 5 minutes
  mu_init[mu_init <= config['mu_min']] = config['mu_min']
  # canonical retention time shouldn't be bigger than largest real RT
  mu_max = dff['retention_time'].max()
  mu_init[mu_init > mu_max] = mu_max

  # take retention times and distort
  if config['rt_distortion'] > 0:
    logger.info('Distorting RTs by {} minutes for initial value generation'.format(config['rt_distortion']))

  rt_distorted = dff['retention_time'] + np.random.normal(0, config['rt_distortion'], len(dff['retention_time']))
  # make sure distorted retention times stay within bounds of real ones
  rt_distorted[rt_distorted > dff['retention_time'].max()] = dff['retention_time'].max()
  rt_distorted[rt_distorted < dff['retention_time'].min()] = dff['retention_time'].min()

  # initialize priors for the segmented linear regression
  # first element of vector is beta_0, or the intercept
  # second element is beta_1 and beta_2, the slopes of the two segments
  beta_init = np.array((
    np.repeat(10, num_experiments), 
    np.repeat(1, num_experiments)), dtype=float)

  logger.info('Optimizing priors with linear approximation for {} iterations.'.format(config['prior_iters']))

  mu_pred = np.zeros(num_peptides)
  # temporary data frame to quickly map over in the loop
  dft = pd.DataFrame(dict(
    stan_peptide_id=dff['stan_peptide_id'], 
    exp_id=dff['exp_id'], 
    pep=dff['pep'], 
    retention_time=mu_init[dff['stan_peptide_id']]))

  for i in range(0, config['prior_iters']):
    # for each experiment, fit a simple linear regression
    # between the distorted RTs and the initial canonical retention times
    for j in range(0, num_experiments):
        idx     = (dff['exp_id'] == j)
        rt_cur  = rt_distorted[idx]
        mu_cur  = mu_init[dff['stan_peptide_id'][idx]]
        pep_cur = dff['pep'][idx]

        # for this experiment, run a linear regression (1 degree polyfit)
        # of the mus and the distorted RTs. store the linear regression params
        m, c = np.polyfit(mu_cur, rt_cur, 1, w=(1 - pep_cur))
        beta_init[(0,1), j] = [c, m]

    # calculate new set of canonical RTs based on linear regression params
    mu_pred = (rt_distorted - beta_init[0][dff['exp_id']]) / beta_init[1][dff['exp_id']] 
    # make sure new canonical RTs are within same range as distorted RTs
    mu_pred[mu_pred <= config['mu_min']] = config['mu_min']
    #mu_pred[mu_pred <= 0] = rt_distorted.min()
    mu_pred[mu_pred >= rt_distorted.max()] = rt_distorted.max()
    dft['retention_time'] = np.copy(mu_pred)

    mu_prev = np.copy(mu_init)

    # new set of priors for canonical RTs based on weighted combination of
    # this set of predicted canonical RTs
    mu_init = dft.groupby('stan_peptide_id')[['pep', 'retention_time']].apply(get_mu).values

    logger.info('Iter {} | Avg. canonical RT shift: {:.5f}'.format(i + 1, 
      pow(np.sum(mu_prev - mu_init), 2) / len(mu_init)))


  # grab linear regression params
  beta_0 = beta_init[0]
  beta_1 = beta_init[1]

  # apply lower bound of (-1.0 * min(beta_1) * min(muInit)) to beta_0
  # where (-1.0 * min(beta_1) * min(muInit)) is the lowest possible intercept
  # given the lowest possible mu and lowest possible beta_1
  beta_0[beta_0 <= (-1 * beta_1.min() * mu_init.min())] = \
    (-1 * beta_1.min() * mu_init.min()) + 1e-3
  # apply lower bound of 0 to beta_1
  beta_1[beta_1 <= 0] = 1e-3
  
  # apply upper bound to prior canonical RTs
  mu_init[mu_init >= dff['retention_time'].max()] = 0.95 * dff['retention_time'].max()

  # create prior list for STAN
  init_list = {
    'mu': mu_init.tolist(),
    #'sigma': sigma_init,
    'beta_0': beta_0.tolist(),
    'beta_1': beta_1.tolist(),
    'sigma_slope': np.repeat(0.1, num_experiments).tolist(),
    'sigma_intercept': np.repeat(0.1, num_experiments).tolist()
  }

  return init_list

def generate_inits_two_piece_linear(dff, config):

  # this is essentially the same process as the linear fit
  init_list = generate_inits_linear(dff, config)

  # set beta_2 (slope of second segment) to the same as the slope of the 1st segment
  beta_2 = np.copy(init_list['beta_1'])
  # apply lower bound of 0 to beta_2
  beta_2[beta_2 <= 0] = 1e-3

  # set beta_2
  init_list['beta_2'] = beta_2.tolist()
  # set split point to be the median canonical RT
  init_list['split_point'] = np.repeat(np.median(init_list['mu']), 
                                       len(init_list['beta_0'])).tolist()

  return init_list

def muij_two_piece_linear(df, mu, params):
  muij = pd.Series(np.zeros(df.shape[0]))
  # if the mu is before the split point, only account for the first segment
  muij[mu <= df['split_point']] = \
    df['beta_0'] + (df['beta_1'] * mu)
  # if the mu is after the split point, account for both segments
  muij[mu  > df['split_point']] = \
    df['beta_0'] + (df['beta_1'] * df['split_point']) + \
    (df['beta_2'] * (mu-df['split_point']))

  return muij

def mu_two_piece_linear(df, mu, params):
  mu_pred = pd.Series(np.zeros(df.shape[0]))

  mu_pred[mu <= df['split_point']] = \
    (df['retention_time'] - df['beta_0']) / df['beta_1']

  mu_pred[mu  > df['split_point']] = \
    ((df['retention_time'] - df['beta_0'] - (df['beta_1']*df['split_point'])) \
    / df['beta_2']) + df['split_point']

  return mu_pred

def muij_linear(exp, exp_id, params):
  muij = params['exp']['beta_0'][exp_id] + \
    params['exp']['beta_1'][exp_id] * exp['mu']
  return muij

def mu_linear(df, params):
  mu_pred = (df['retention_time'] - df['beta_0']) / df['beta_1']
  return mu_pred

def sigmaij_linear_mu(df, params):
  # get sigmaij from the sigma_intercept and sigma_slope parameters for this experiment
  sigmaij = df['sigma_intercept'] + ((df['sigma_slope'] / 100) * df['mu'])
  return sigmaij

# uniform RT density from min(RT) to max(RT)
def uniform_null(exp):
  return 1 / (np.max(exp['retention_time']) - np.min(exp['retention_time']))

# normal RT density, truncated at min(RT) and max(RT)
def normal_null(exp):
  rt_mean = np.mean(exp['retention_time'])
  rt_std = np.std(exp['retention_time'])
  return norm.pdf(exp['retention_time'], loc=rt_mean, scale=rt_std)

def mixture_normal_normal(exp):
  # Fit3d - mixture between two normal densities
  
  rt_mean = np.mean(exp['retention_time'])
  rt_std = np.std(exp['retention_time'])

  comp1 = exp['pep'] * \
    norm.pdf(exp['retention_time'], loc=rt_mean, scale=rt_std)
  comp2 = (1.0 - exp['pep']) * \
    norm.pdf(exp['retention_time'], loc=exp['muij'], scale=exp['sigmaij'])

  return comp1 + comp2

def normal_drt(exp):
  return norm.pdf(exp['retention_time'], loc=exp['muij'], scale=exp['sigmaij'])

def laplace_drt(exp):
  return laplace.pdf(exp['retention_time'], loc=exp['muij'], scale=exp['sigmaij'])

models = {
  'linear': {
    'name': 'linear',
    # model names must be valid C++ class names. no dashes, etc.
    'model_name': 'FitLinear',
    # name of the executable file-folder in the models/ folder
    'model_binary': 'fit_RT_linear',
    # function to generate initial values
    # should return dict of initial values, where names of initial values
    # are the same as parameter names.
    # values should be in vanilla python list format, 
    # not numpy arrays or pandas series
    'init_func': generate_inits_linear,
    # when serializing the alignment results, which values to store in which file
    # these should be the same name as the STAN parameter names
    'exp_keys': ['beta_0', 'beta_1', 'sigma_intercept', 'sigma_slope'],
    'pair_keys': ['muij', 'sigma_ij'],
    'peptide_keys': ['mu'],
    # transform reference RT from reference space to absolute RT space
    'ref_to_rt': muij_linear,
    # inverse of above. transform absolute RTs into reference space
    'rt_to_ref': mu_linear,

    'sigmaij_func': sigmaij_linear_mu,
    # function for determining density over all RTs in experiment
    # P(RT|PSM-)
    'rt_minus_func': normal_null,
    # function for likelihood of RT given correct assignment
    # P(RT|PSM+)
    #'rt_plus_func': mixture_normal_normal
    'rt_plus_func': normal_drt
  },
  'two_piece_linear': {
    'name': 'two_piece_linear',
    'model_name': 'FitTwoPieceLinear',
    'model_binary': 'fit_RT3e',
    'init_func': generate_inits_two_piece_linear,
    'exp_keys': ['beta_0', 'beta_1', 'beta_2', 
      'split_point', 'sigma_intercept', 'sigma_slope'],
    'pair_keys': ['muij', 'sigma_ij'],
    'peptide_keys': ['mu'],
    'ref_to_rt': muij_two_piece_linear,
    'rt_to_ref': mu_two_piece_linear,
    'sigmaij_func': sigmaij_linear_mu,
    'rt_minus_func': normal_null,
    'rt_plus_func': normal_drt
  },
  'two_piece_linear_laplace': {
    'name': 'two_piece_linear_laplace',
    'model_name': 'FitTwoPieceLinearLaplace',
    'model_binary': 'fit_RT3e',
    'init_func': generate_inits_two_piece_linear,
    'exp_keys': ['beta_0', 'beta_1', 'beta_2', 
      'split_point', 'sigma_intercept', 'sigma_slope'],
    'pair_keys': ['muij', 'sigma_ij'],
    'peptide_keys': ['mu'],
    'ref_to_rt': muij_two_piece_linear,
    'rt_to_ref': mu_two_piece_linear,
    'sigmaij_func': sigmaij_linear_mu,
    'rt_minus_func': normal_null,
    'rt_plus_func': laplace_drt
  }
}

def get_model_from_config(config):
  model = 'two_piece_linear'
  if config['model'] is not None:
    if config['model'] in models:
      model = config['model']
    else:
      raise ConfigFileError('Model \"{}\" not found. Available choices are: {}'.format(
        model, models.keys()))
  else:
    logger.info('Alignment model not defined. Defaulting to \"two_piece_linear\" model')

  return models[model]




#!/usr/bin/env python3
# coding: utf-8

import argparse
import json
import logging
import numpy as np
import os
import pandas as pd
import pkg_resources
import sys
import time

from dart_id.align import align
from dart_id.converter import process_files
from dart_id.figure_gen import exp_alignment, residuals, newpeps
from dart_id.helper import add_global_args, read_config_file, init_logger, load_params_from_file
from shutil import copyfile
from string import Template

logger = logging.getLogger('root')

def figures(df, config=None, params=None):

  logger.info('Saving figures to {}'.format(config['output']))

  fig_data = {}

  try:
    fig_data['alignment'] = exp_alignment.gen(df, config, params, config['output'])
  except: pass

  try:
    fig_data['residuals'] = residuals.gen(df, config, params, config['output'])
  except: pass
  
  try:
    fig_data['newpeps'] = newpeps.gen(df, config, params, config['output'])
  except: pass

  generate_html(fig_data, config['output'])

  logger.info('Figure generation complete')

def generate_html(fig_data, output_path):
  logger.info('Generating HTML...')

  # build dict for variable injection into the HTML file
  fig_data = { 'data': json.dumps(fig_data) }

  # load HTML template and inject data variable
  logger.info('Loading template HTML and injecting variables...')
  template = pkg_resources.resource_string('dart_id', '/'.join((
    'figure_resources', 'template.html')))
  template = template.decode('utf-8')
  template = Template(template).safe_substitute(fig_data)

  # write HTML template to output directory
  html_out = os.path.join(output_path, 'figures.html')
  logger.info('Writing template HTML file to {}'.format(html_out))
  with open(html_out, 'w') as template_out:
    template_out.write(template)

  # move resource files (CSS and JS)
  logger.info('Moving HTML resource files')
  resource_files = [
    'bootstrap.min.css', 'bootstrap.min.js', 
    'jquery-3.3.1.slim.min.js', 'styles.css'] #,
    #'jquery.dataTables.min.css', 'jquery.dataTables.min.js'
  #]
  for i in resource_files:
    copyfile(pkg_resources.resource_filename(
      'dart_id', '/'.join(('figure_resources', i))), 
    os.path.join(output_path, 'figures', i))


def main():
  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser)
  args = parser.parse_args()

  # load config file
  # this function also creates the output folder
  config = read_config_file(args)

  # initialize logger
  init_logger(config['verbose'], os.path.join(config['output'], 'figures.log'))

  # load first input file and replace home and user vars
  f = config['input'][0]
  f = os.path.expanduser(f)
  f = os.path.expandvars(f)

  # read in input files
  df = pd.read_csv(f, sep='\t', low_memory=False)
  params = load_params_from_file(config['params_folder'])

  # TODO: check that the columns we need are present

  # generate figures
  figures(df, config=config, params=params)

if __name__ == '__main__':
  main()


#!/usr/bin/env python3
# coding: utf-8

import logging
import matplotlib
matplotlib.use('PS')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from dart_id.helper import *
from scipy.stats import norm, lognorm, laplace, gaussian_kde

logger = logging.getLogger('root')

def gen(df, config, params, output_path):
  figures_path = create_fig_folder(output_path, 'figures')
  fig_names = []
  col_names = config['col_names']

  # PEP vs. PEP.new scatterplot
  inds = (~pd.isnull(df['pep_new'])) & (df['pep_new'] > 1e-5) & (df[col_names['pep']] > 1e-5)

  f, ax = plt.subplots()
  hst = ax.hist2d(np.log10(df[col_names['pep']][inds]), np.log10(df['pep_new'][inds]), bins=(50, 50), cmap=plt.cm.Reds)
  ax.plot([-5, 0], [-5, 0], '-r')

  ax.grid()

  interval = (-5, 0)
  ax.set_xlim(interval)
  ax.set_ylim(interval)
  ax.set_xlabel('Spectral PEP', fontsize=16)
  ax.set_ylabel('DART-ID PEP', fontsize=16)
  interval = np.arange(-5, 1, 1)
  ax.set_xticks(interval)
  ax.set_yticks(interval)
  ax.set_xticklabels(['$10^{{{}}}$'.format(i) for i in interval], fontsize=12)
  ax.set_yticklabels(['$10^{{{}}}$'.format(i) for i in interval], fontsize=12)

  f.set_size_inches(7, 6)
  # plt.tight_layout()

  cbar = plt.colorbar(hst[3], ax=ax)
  cbar.set_label('Frequency', fontsize=16, labelpad=20, ha='center', va='top')
  cbar.ax.xaxis.set_label_position('top')

  fname = os.path.join(figures_path, 'pep_new_scatterplot.png')
  logger.info('Saving figure to {} ...'.format(fname))
  f.savefig(fname, dpi=160)
  fig_names.append(fname)

  plt.close()
  f.clf()

  # Fold-change increase
  num_points=100
  x = np.logspace(-5, 0, num=num_points)
  y = np.zeros(num_points)
  y2 = np.zeros(num_points)
  y3 = np.zeros(num_points)
  inds = ~pd.isnull(df['pep_new'])

  for i, j in enumerate(x):
      y[i] = np.sum(df['dart_PEP'] < j) / np.sum(df[col_names['pep']] < j)
      y2[i] = np.sum(df[col_names['pep']] < j) / df.shape[0]
      y3[i] = np.sum(df['dart_PEP'] < j) / df.shape[0]

  f, (ax1, ax2) = plt.subplots(2, 1)

  ax1.semilogx(x, (y*100)-100, '-b')
  # ax1.plot([np.min(x), np.max(x)], [0, 0], '-r', linestyle='dashed', linewidth=2)
  ax1.plot([1e-2, 1e-2], [-1000, 1000], '-k', linestyle='dashed')
  ax1.grid()
  ax1.set_xlim([3e-4, 3e-1])
  ax1.set_ylim([-25, np.max(y)*100-50])
  ax1.set_xlabel('PEP Threshold', fontsize=16)
  ax1.set_ylabel('Increase (%)', fontsize=16)
  ax1.set_title('Increase in confident PSMs', fontsize=16)

  ax2.semilogx(x, y2, '-b', linewidth=1, label='Spectra PEP')
  ax2.semilogx(x, y3, '-g', linewidth=1, label='DART-ID PEP')
  #ax2.fill_between(x, 0, y2)
  ax2.plot([1e-2, 1e-2], [-1000, 1000], '-k', linestyle='dashed')
  ax2.grid()
  ax2.set_xlim([3e-4, 3e-1])
  ax2.set_ylim([0, 1.05])
  ax2.set_xlabel('PEP Threshold', fontsize=16)
  ax2.set_ylabel('Fraction', fontsize=16)
  ax2.set_title('Fraction of confident PSMs', fontsize=16)
  ax2.legend(fontsize=16)
  plt.subplots_adjust(hspace=0.6, wspace=0.3)

  f.set_size_inches(5, 7)
  # plt.tight_layout()

  fname = os.path.join(figures_path, 'fold_change_ids.png')
  logger.info('Saving figure to {} ...'.format(fname))
  f.savefig(fname, dpi=160)
  fig_names.append(fname)

  plt.close()
  f.clf()

  # MS1 Intensity CV Validation
  '''
  prots = df[col_names['leading_protein']]

  # get the contaminant and decoy filter tags, if they exist
  filter_con = next((f for f in config['filters'] if f['name'] == 'contaminant'))
  if filter_con is not None: filter_con = filter_con['tag']
  else:                      filter_con = '!'
  filter_rev = next((f for f in config['filters'] if f['name'] == 'decoy'))
  if filter_rev is not None: filter_rev = filter_rev['tag']
  else:                      filter_rev = '!'

  prots = prots.loc[(~prots.str.contains(filter_con)) & 
                    (~prots.str.contains(filter_rev))]
  prots.reset_index(drop=True)

  prot_list = prots.value_counts()
  prot_list = prot_list.loc[prot_list > 50]

  pep_thresh = 0.05

  cvs = np.zeros((len(prot_list), 3))
  n_cvs = np.zeros((len(prot_list), 3))

  for i, j in enumerate(prot_list):
      prot_name = prot_list.index[i]
      intensities = df[col_names['intensity']][(df[col_names['leading_protein']] == prot_name) & 
                                    (~pd.isnull(df[col_names['intensity']])) & 
                                    (df[col_names['pep']] < pep_thresh)]
      cvs[i][0] = np.std(intensities) / np.mean(intensities)
      n_cvs[i][0] = len(intensities)
      
      intensities = df[col_names['intensity']][(df[col_names['leading_protein']] == prot_name) & 
                                    (~pd.isnull(df[col_names['intensity']])) & 
                                    (df['pep_new'] < pep_thresh) &
                                    (df[col_names['pep']] > pep_thresh)]
      cvs[i][1] = np.std(intensities) / np.mean(intensities)
      n_cvs[i][1] = len(intensities)
      
      dfa = df.loc[(~pd.isnull(df[col_names['intensity']])) & 
                   (df[col_names['pep']] < pep_thresh) &
                   (~df[col_names['leading_protein']].str.contains(filter_con)) & 
                   (~df[col_names['leading_protein']].str.contains(filter_rev))
                  ].sample(n=j)
      cvs[i][2] = np.std(dfa[col_names['intensity']]) / np.mean(dfa[col_names['intensity']])
      n_cvs[i][2] = dfa.shape[0]

  # remove pairs which have NAs in them
  na_rows = np.apply_along_axis((lambda x: np.isnan(x).any()), 1, cvs)
  na_rows = na_rows | (np.apply_along_axis((lambda x: (x == 1).any()), 1, n_cvs))
  cvs = cvs[~na_rows]
  n_cvs = n_cvs[~na_rows]

  x = np.linspace(0, 7.5, 1000)
  cvs_density = gaussian_kde(cvs[:,0])
  new_cvs_density = gaussian_kde(cvs[:,1])
  null_cvs_density = gaussian_kde(cvs[:,2])

  f, ax = plt.subplots()

  ax.plot(x, cvs_density(x), '-b', label='Spectra PEP < {}'.format(pep_thresh))
  ax.plot(x, new_cvs_density(x), '-g', label='Spectra+RT PEP < {}'.format(pep_thresh))
  ax.plot(x, null_cvs_density(x), '-r', label='Null')
  ax.set_xlabel('Intra-Protein MS1 Intensity CV ($\sigma$/$\mu$)', fontsize=16)
  ax.set_ylabel('Density', fontsize=16)
  ax.set_title('Intra-Protein MS1 Intensity CV\n Validation of Newly Boosted Observations', fontsize=16)
  ax.tick_params(labelsize=16)
  ax.legend(fontsize=16)

  f.set_size_inches(7, 7)
  plt.tight_layout()

  fname = os.path.join(figures_path, 'ms1_intensity_validation.png')
  logger.info('Saving figure to {} ...'.format(fname))
  f.savefig(fname, dpi=160)
  fig_names.append(fname)

  plt.close()
  f.clf()
  '''

  return fig_names




#!/usr/bin/env python3
# coding: utf-8

import argparse
import importlib
import logging
import numpy as np
import os
import pandas as pd
import re

from dart_id.helper import *

logger = logging.getLogger('root')

def main():
  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser, add_config_file=False)

  parser.add_argument('--include', type=str, default=None, help='Inclusion expression (regular expression) that is evaluated against all raw file names to determine whether or not that raw file is included.')
  parser.add_argument('--raw-file-col', type=str, default='Raw file', help='Name of the raw file column.')
  args = parser.parse_args()

  # initialize logger
  init_logger(args.verbose, '', log_to_file=False)
  logger = logging.getLogger('root')

  df = pd.DataFrame()

  # iterate through each input file provided.
  for i, f in enumerate(args.input):

    logger.info('Reading in input file #{} | {} ...'.format(i+1, f.name))

    # load the input file with pandas
    # 
    # have a variable low memory option depending on the input type.
    # MaxQuant, for example, has a structure that forces pandas out of its
    # optimal low memory mode, and we have to specify it here.
    dfa = pd.read_csv(f.name, sep='\t', low_memory=False)

    # keep track of where observations came from. this is _not_ the raw file ID
    # but instead the ID from which input file it originated from, so that if
    # we need to split these observations up by input file in the future we can do so
    dfa['input_id'] = i

    # append a copy of dfa into df, because the conversion process will heavily
    # modify dfa. we need to keep a copy of the original dataframe in order to append
    # the new columns back onto it later.
    # re-index columns with '[dfa.columns.tolist()]' to preserve the general column order
    df = df.append(dfa)[dfa.columns.tolist()]

  logger.info('Raw files:')
  logger.info(df[args.raw_file_col].unique())

  if args.include is not None:
    if len(args.include) == 0:
      raise Exception('Experiment inclusion expression by raw file name provided, but expression is defined incorrectly.')
    else:
      # see if any raw file names match the user-provided expression
      include_exps = list(filter(lambda x: re.search(r'' + args.include + '', x), df[args.raw_file_col].unique()))

      logger.info('{} raw files match the inclusion expression \"{}\"'.format(len(include_exps), args.include))
      logger.info(include_exps)
      logger.info('Retaining {} observations out of {} total'.format(np.sum(df[args.raw_file_col].isin(include_exps)), df.shape[0]))

      include_exps = df[args.raw_file_col].isin(include_exps).values
      df = df.loc[include_exps].reset_index(drop=True)

  # if combining input files, then write to one combined file
  logger.info('Combining input file(s) and writing adjusted data file to {} ...'.format(args.output))
  df.to_csv(args.output, sep='\t', index=False)


if __name__ == '__main__':
  main()